#include "batch.h"

Batch::Batch(const std::list<std::string>& batch_list)
    : _batchList(batch_list),
    _cmed(ClusterMediator::Instance()), _isRunning(false),_thread(nullptr)
{
}
Batch::~Batch()
{
    if (_thread) _thread->detach();
}

void Batch::Cancel()
{
    _isRunning = false;
}

void Batch::Run()
{
    _isRunning = true;
    _thread = new std::thread( std::bind(&Batch::Process, this) );
}

void Batch::Process()
{
    try
    {
        //Must do this *before* the loop starts!
        CluSpecPtr batch_clus_spec = _cmed.GetCluSpecPtr(wv::DEFAULT_CID)->MintNew();
        if ( !batch_clus_spec->ClusterParameters() )
        {
            wv::CLUSTER_METHOD method = batch_clus_spec->ClusterMethod();
            batch_clus_spec->SetClusterParameters( ClusterParamsBase::Create(method) );
        }
        batch_clus_spec->ClusterParameters()->SetSampleRangePars();

        //Now do all the clustering
        for (auto& it : _batchList)
        {
            if (!_isRunning) break;

            WavesFilePtr waves_file = WavesFilePtr(WaveformsInBase::Create(it)); //true for _isBatch
            waves_file->SetIsBatch(true);
            try
            {
                int num_waveforms = waves_file->Load();
                if (num_waveforms==0)
                    throw WIOException("Batch::Process", it,
                                      "ERROR: File was not opened or read properly, no waveforms loaded.");
            }
            catch (WIOException& e)
            {
                Log(e.what());
                continue;
            }

            //We create the list of cids from the electrodes in the data file.  Thus each ClusterMgr
            //is dedicated to all the electrodes in a single file.
            std::vector<wv::CLUSTER_ID> cids;
            for (auto it : waves_file->Electrodes())
            {
                waves_file->GenerateTransforms(it);
                //Set up the wave data for the given electrode
                wv::CLUSTER_ID cid = _cmed.IncrementCID();
                CluSpecPtr new_clus_spec = batch_clus_spec->MintNew();
                cids.push_back(cid);
                _cmed.AddClusterSpec(cid, new_clus_spec);
                _cmed.AddInputFile(cid, waves_file);
                _cmed.SetElectrode(cid, it);
            }

            ClusterMgrPtr cmgr( new ClusterMgr(cids) );
            _cmed.AddClusterMgr(cmgr);
            cmgr->StartThreads();
            cmgr->Wait();
        }
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }

    _batchList.clear();
    emit BatchComplete();
}
