#include "clusmethodtobase.h"
#include "manualto.h"
#include "autoclassto.h"
#include "kmeansto.h"
#include "spc_to.h"
#include "klustakwikto.h"
#include "../../memrep/parameters/cluster/clusterparamsbase.h"
#include "../../memrep/clustermediator.h"

ClusMethodTOBase* ClusMethodTOBase::Create(wv::CLUSTER_ID cid, wv::TRANSFORM transform, wv::CLUSTER_METHOD cluster_method,
                            MatCPtr data, const std::string& wfm_file)
{
    ClusMethodTOBase* to(NULL);
    switch (cluster_method)
    {
        case wv::MANUAL:
            break;
        case wv::AUTOCLASS:
            to = new AutoClassTO(cid, transform, data, wfm_file);
            break;
        case wv::KMEANS:
            to = new KMeansTO(cid, transform, data, wfm_file);
            break;
        case wv::SPC:
            to = new SPC_TO(cid, transform, data, wfm_file);
            break;
        case wv::KK:
            to = new KlustaKwikTO(cid, transform, data, wfm_file);
            break;
        case wv::UNKNOWN_METHOD:
            throw BadClusterMethod("ClusMethodTOBase::Create", cluster_method);
    }
    return to;
}
ClusMethodTOBase* ClusMethodTOBase::Create(wv::CLUSTER_ID cid,wv::CLUSTER_METHOD cluster_method,
                            WaveData* wave_data, const std::string& wfm_file)
{
    ClusMethodTOBase* to(NULL);
    switch (cluster_method)
    {
        case wv::MANUAL:
            to = new ManualTO(cid, wave_data, wfm_file);
                // the wfm_file is still used to determine what the file name, transform name is
                //Need to send wavedata in this case, since the polylines can be based of *any* set of
                //transforms
            break;
        case wv::AUTOCLASS:
        case wv::KMEANS:
        case wv::SPC:
        case wv::KK:
        case wv::UNKNOWN_METHOD:
            throw BadClusterMethod("ClusMethodTOBase::Create", cluster_method);
            break;
    }
    return to;
}

ClusMethodTOBase::ClusMethodTOBase(wv::CLUSTER_ID cid, wv::TRANSFORM transform, WaveData* wave_data, int num_records,
                                   const std::string& wfm_file)
    : AnalysisTOBase(wave_data, num_records, wfm_file), _cid(cid), _transform(transform), _clusProc(NULL), _mcluFile(NULL)
{
    Setup();
}
ClusMethodTOBase::ClusMethodTOBase(wv::CLUSTER_ID cid, wv::TRANSFORM transform, MatCPtr data, int num_records,
                                   const std::string& wfm_file)
    : AnalysisTOBase(data, num_records, wfm_file), _cid(cid), _transform(transform), _clusProc(NULL), _mcluFile(NULL)
{
    Setup();
}
ClusMethodTOBase::~ClusMethodTOBase()
{
    if (_mcluFile) delete _mcluFile;
    if (_clusProc && _clusProc->isOpen())
    {
        _clusProc->kill();
        delete _clusProc;
    }
}

void ClusMethodTOBase::SetWindow(int window_begin, int window_end)
{
    _windowBegin = window_begin;
    _windowEnd = window_end;
}

void ClusMethodTOBase::LoadClassification(int exit_code)
{
    try
    {
        Log(Name() + " analysis completed in " + std::to_string(TimeElapsed()/1000.0)
            + " seconds with exit code " + std::to_string(exit_code) + ".");

        //Read the .clu file into a Classification, setting -1 for unclassified waveforms
        boost::dynamic_bitset<> window( NumRecords() );
        ut::FillBitset(window, true, WindowBegin(), WindowEnd() );
        std::string mclu_file_name = FileBase::MakeTempFileName(FileBase::UserPath(), InputFileName(), Transform(), ID(), "mclu", Electrode());
        if (_mcluFile) delete _mcluFile; //Should not actually happen
        _mcluFile = new MCluFileIn(mclu_file_name, window);

        _mcluFile->Load();
        _mcluFile->Close();

        const Classification* cls = _mcluFile->CData();
        if (!cls)
            Log("ERROR: Could not open " + mclu_file_name + " analysis not recorded.");
        else if (cls->VClusters().size() == 0)
            cls = NULL;

        //Update ClusterMediator and finish
        ClusterMediator& cmed = ClusterMediator::Instance();
        cmed.AddClustering( _cid, ClassificationCPtr(cls) );
        wv::CLUSTER_ID* pcid = new wv::CLUSTER_ID(_cid);

        //Now calculate statistics for the best level.  The algorithm is encapsulated in a class in case we ever want the user to be able
        //to select between different measures.
        int level = cls->BestLevel();
        wv_vector_int* vcluster = cls->VClusters().at(level).second;
        CluSpecPtr clu_spec = cmed.GetCluSpecPtr(_cid);

        bool is_manual = Name()=="Manual";
            //Note there is no principled way to construct a specific transform for Manual clustering--
            //the circles, polylines etc. can come from any combination of transforms.  So will just use raw.
        wv::TRANSFORM waves_transform = clu_spec->GetWSGSpec().WavesTransform();

        int j_begin, j_end;
        if (is_manual && waves_transform!=wv::RAW)
        {
            j_begin = 0;
            j_end = WData()->NumSamples(wv::RAW, wv::DEFAULT_TID);
        }
        else
        {
            j_begin = clu_spec->SelectedSamplesBegin();
            j_end = clu_spec->SelectedSamplesEnd();
        }
            //That is, calculate the M-Index off the selected sample range if it's an automatic clustering
            //or if it's manual and the user has selected a range from wv::RAW in the gui.

        Dunn dunn( is_manual ? WData()->Waveforms(wv::RAW) : Waves(),
                    vcluster, cls->Clusters(level), j_begin, j_end);
        dunn.Calculate();
        std::vector<wv::float_t> dunn_indexes = dunn.Indexes();
        cmed.SetStats(_cid, level, dunn_indexes);

        //Finish
        AnalysisFinished(pcid);
            //Doing this hack with a pointer so we can keep using the same interface as with transforms, where
            //a pointer is sent
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}
void ClusMethodTOBase::AnalysisFinished(void* classification)
{
    AnalysisTOBase::AnalysisFinished(classification);
}
void ClusMethodTOBase::AnalysisFinished(const void* classification) const
{
    AnalysisTOBase::AnalysisFinished(classification);
}
void ClusMethodTOBase::Cluster(QProcess* clus_proc, const StringList& proc_args)
{
    _clusProc = clus_proc;
    std::string proc_name = Name(),
#if defined(Q_WS_X11) || defined(Q_OS_LINUX)
            proc_path = QApplication::applicationDirPath().toStdString()
                    + "/ClusProgs/" + proc_name + "/" + proc_name;
#elif defined Q_WS_MAC || defined Q_OS_MACX
            proc_path = QApplication::applicationDirPath().toStdString()
            + "/../../../ClusProgs/" + proc_name + "/" + proc_name + ".app";
#elif defined Q_OS_WIN
            proc_path = QApplication::applicationDirPath().toStdString()
                    + "/../ClusProgs/" + proc_name + "/release/" + proc_name;
#endif
    connect(_clusProc,SIGNAL(readyReadStandardOutput()), this, SLOT(ReadStandardOutput()));
    connect(_clusProc,SIGNAL(readyReadStandardError()), this, SLOT(ReadStandardError()));
    connect(_clusProc,SIGNAL(stateChanged(QProcess::ProcessState)), this, SLOT(StateChange(QProcess::ProcessState)));
    connect(_clusProc,SIGNAL(error(QProcess::ProcessError)), this, SLOT(Error(QProcess::ProcessError)));
    connect(_clusProc,SIGNAL(finished(int)),this,SLOT(LoadClassification(int)));
    _clusProc->start(proc_path.c_str(), ut::ToQStringList(proc_args));
    QApplication::processEvents();
}

void ClusMethodTOBase::ReadStandardOutput()
{
    std::string out = _clusProc->readAllStandardOutput().data();
    Log(Name() + " stdout: " + out);
}
void ClusMethodTOBase::ReadStandardError()
{
    std::string err = _clusProc->readAllStandardError().data();
    Log(Name() + " stderr: " + err);
}
void ClusMethodTOBase::StateChange(QProcess::ProcessState st)
{
    Log( Name() + " new state: " + std::to_string(st) );
}
void ClusMethodTOBase::Error(QProcess::ProcessError)
{
    Log( Name() + " error: " + _clusProc->errorString().toStdString() );
}

std::string ClusMethodTOBase::InputPathStub()
{
    StringVec elts = ut::Split<StringVec>(InputFullPath(), ".");
    elts.pop_back();
    elts.pop_back();
    std::string stub = boost::algorithm::join(elts, ".");
    return stub;
}

std::string ClusMethodTOBase::InputFileName() const
{
    std::string dir, file_name;
    wv::CLUSTER_ID cid;
    wv::TRANSFORM transform;
    int electrode;
    FileBase::TempFileComponents(InputFullPath(), dir, file_name, transform, cid, electrode);
        //Lots of unnecessary information but it's cleaner this way; the input to cluster methods are
        // *always* temporary files.
    return file_name;
}
void ClusMethodTOBase::Setup()
{
    qRegisterMetaType<QProcess::ProcessState>("QProcess::ProcessState");
    qRegisterMetaType<QProcess::ProcessError>("QProcess::ProcessError");

    wv::CLUSTER_ID dummy;
    wv::TRANSFORM dummyt;
    FileBase::TempFileComponents(InputFullPath(), _userDir,
                                 _shortInputFileName, dummyt, dummy, _electrode);

    CluSpecPtr clu_spec = ClusterMediator::Instance().GetCluSpecPtr( ID() );
    SetParams( clu_spec->ClusterParameters() );
}
