#include "kmeansto.h"
#include "../../memrep/clustermediator.h"

KMeansTO::KMeansTO(wv::CLUSTER_ID cid, wv::TRANSFORM transform, MatCPtr waves, const std::string& wfm_file) :
        ClusMethodTOBase(cid, transform, waves, waves->size1, wfm_file)
{
//    MakeProgressBar(Name());
}

void KMeansTO::Start()
{
    StartTimer();

    //Get the parameters
    StringList proc_args;
    proc_args << InputPathStub();
    proc_args << std::to_string( ID() );
    proc_args << std::to_string( Electrode() );
    proc_args << std::to_string( WindowBegin() )
            << std::to_string( WindowEnd() );
        //The subset of records actually being used in the clustering.  Use fseek

    int num_pars = Params()->NumPars();
    for (int i=0; i<num_pars; ++i)
        proc_args << Params()->Value(i); //NB Min # clusters == 1
    int samples_begin = boost::lexical_cast<int>( Params()->Par("samples_begin") ),
            samples_end = boost::lexical_cast<int>( Params()->Par("samples_end") );
    _numSamplesUsed = samples_end - samples_begin;

    QProcess* clus_proc = new QProcess(); //Must create here, so that it is in this thread
    Cluster(clus_proc, proc_args);
}

void KMeansTO::LoadClassification(int exit_code)
{
    try
    {
        //Initialize number of clusters
        int k_max = boost::lexical_cast<int>(Params()->Par("Max # of clusters"));
        _numClusterings = k_max>1 ? k_max : 2;

        //Read the KMeans output file into memory
        KMeansFileIn km_file(InputFullPath() + "." + std::to_string(ID()) + ".km", _numClusterings,
                             NumRecords(), _numSamplesUsed);
        int num_records = km_file.Load();
        if (num_records == 0) return;
        const std::map<int, KMPair>* clusters = km_file.KMData();
        int best_k = MaxJump(clusters);

        //Convert the KMeans output structure in to a std::list<Clustering>
        std::vector<ConstClustering> clusterings;
        for (auto it : *clusters)
        {
            int level = it.first;
            const wv_vector_int* clustering = it.second.first;
            ConstClustering clus_pair(level, clustering);
            clusterings.push_back(clus_pair);
        }

        //Write the best cluster output to a .mclu file
        std::string mclu_file_name = FileBase::MakeTempFileName(FileBase::UserPath(), InputFileName(), Transform(), ID(), "mclu", Electrode());
        MCluFileOut mclu_file(mclu_file_name, best_k, clusterings);
        mclu_file.Save();
        mclu_file.Close(); //Make sure to close before reading this in with LoadClassification!

        //Then read the .mclu file in the normal way
        ClusMethodTOBase::LoadClassification(exit_code);
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

wv::float_t KMeansTO::Distortion(const wv_vector_int* assignment, const wv_matrix* centers, int k)
{
    if (k==0) return 0;

    const wv_matrix* waves = Waves().get();
    int num_records = waves->size1,
            num_total_samples = waves->size2,
            samples_begin = boost::lexical_cast<int>( Params()->Par("samples_begin") );

    //Create a submatrix with just the samples used
    wv_matrix* subwaves = wv_matrix_alloc(num_records, _numSamplesUsed);
    if (num_total_samples == _numSamplesUsed)
        wv_matrix_memcpy(subwaves, waves);
    else
        for (int i=0; i<num_records; ++i)
        {
            wv_vector_const_view row = wv_matrix_const_row(waves, i),
                    subrow = wv_vector_const_subvector(&row.vector, samples_begin, _numSamplesUsed);
            wv_matrix_set_row(subwaves, i, &subrow.vector);
        }

    //Covariance matrix
    wv_matrix* cov_mat = wv_matrix_alloc(_numSamplesUsed, _numSamplesUsed),
            * inv_cov = wv_matrix_alloc(_numSamplesUsed, _numSamplesUsed);
#ifdef USE_GSL
    mgsl::CovarianceP(subwaves, cov_mat);
    gsl_permutation* perm = gsl_permutation_alloc(_numSamplesUsed);
    int signum;
    gsl_linalg_LU_decomp(cov_mat, perm, &signum); //Converts cov_mat to an LU matrix
    gsl_linalg_LU_invert(cov_mat, perm, inv_cov);
#else
    cov_mat = msl::Covariance(subwaves);
    int M = cov_mat->size1, N = cov_mat->size2,
            NI = M<N ? M : N;
    int* ipiv = new int[NI];
    msl::lapack_getrf(LAPACK_ROW_MAJOR, M, N, cov_mat->data, cov_mat->tda, ipiv);
    msl::lapack_getri(LAPACK_ROW_MAJOR, NI, cov_mat->data, cov_mat->tda, ipiv);
    inv_cov = cov_mat; //cov_mat is inverted *in place* with this algorithm
#endif

    wv_vector* m_dist = wv_vector_alloc(k);
    for (int c=0; c<k; ++c) // *** Parallelize!
    {
        //Create a new matrix for the given cluster and put all records that have been
        //assigned to that cluster in it.
        wv_matrix* X = wv_matrix_alloc(num_records, _numSamplesUsed);
            //Invariably larger than needed, manually set size1 to avoid using leftover space
        int ct=0;
        for (int i=0; i<num_records; ++i)
        {
            int c_i = wv_vector_int_get(assignment, i);
            if (c_i == c)
            {
                wv_vector_const_view row = wv_matrix_const_row(subwaves, i);
                wv_matrix_set_row(X, ct, &row.vector);
                ++ct;
            }
        }
        X->size1 = ct; // ***

        //Now compute the Mahalanobis distance.
        wv_vector_const_view center = wv_matrix_const_row(centers, c);
        wv_vector* distances = wv_vector_alloc(ct);
        wv_vector* y = wv_vector_alloc(_numSamplesUsed); //Declare here to avoid reallocating every iteration
        for (int i=0; i<ct; ++i)
        {
            //X - c_X
            wv_vector_view xview = wv_matrix_row(X, i);
            wv_vector* x = &xview.vector;
            wv_vector_sub(x, &center.vector);

            // (X - c_X)' * inv_cov * (X - c_X)
            wv::float_t d;
#ifdef USE_GSL
            gsl_blas_dgemv(CblasTrans, 1.0, inv_cov, x, 0.0, y);
            gsl_blas_ddot(y, x, &d);
#else
            msl::blas_gemv(CblasRowMajor, CblasTrans,
                        inv_cov->size1, inv_cov->size2,
                        (wv::float_t)1.0,
                        inv_cov->data, inv_cov->tda,
                        x->data, x->stride,
                        (wv::float_t)0.0,
                        y->data, y->stride);
            d = msl::blas_dot(y->size, y->data, y->stride, x->data, x->stride);
#endif
            wv_vector_set(distances, i, d);
        }

        //Get the mean Mahalanobis distance for records from this cluster and store.
        wv::float_t dcavg = wv_stats_mean(distances->data, distances->stride, distances->size);
        wv_vector_set(m_dist, c, dcavg);

        wv_vector_free(distances);
        X->size1 = num_records; //So everything gets freed!
        wv_matrix_free(X);
        wv_vector_free(y);
    }

    //Now finally the distortion:
    wv::float_t distortion = (1.0/_numSamplesUsed) * wv_vector_min(m_dist);

    wv_matrix_free(cov_mat);
#ifdef USE_GSL
    wv_matrix_free(inv_cov);
    gsl_permutation_free(perm);
#endif
    wv_vector_free(m_dist);

    return distortion;
}

int KMeansTO::MaxJump(const std::map<int, KMPair>* clusters)
{
    assert(clusters->size()==_numClusterings);
    int num_samples = (int)Waves()->size2,
        expt = -num_samples/2; //So we're tacitly rounding down, currently

    wv::float_t* D = new wv::float_t[_numClusterings+1], // The +1 is for the 0 cluster
            * raw_distortion = new wv::float_t[_numClusterings+1],
            * J = new wv::float_t[_numClusterings],
            * raw_jump = new wv::float_t[_numClusterings];
    Log("Calculating best cluster for K-Means analysis\n");
    std::string distortion_msg = "Distortions by cluster:\n";
    raw_distortion[0] = 1;
    D[0] = 1;
        //Wikipedia says this should be 0, but sample runs clearly indicate that 1 is the limit
    int ct = 1;
    for (auto it : *clusters)
    {
        int k = it.first;
        wv_vector_int* assignment = it.second.first;
        wv_matrix* centers = it.second.second;
        wv::float_t d = Distortion(assignment, centers, k);
        raw_distortion[ct] = d;
        distortion_msg += std::to_string(k) + ", " + std::to_string(d) + "\n";
#ifdef USE_GSL
        D[ct] = gsl_pow_int(d, expt);
#else
        D[ct] = std::pow(d, (wv::float_t)expt);
#endif
//        qDebug() << "D: " << D[ct];
        if (ct>0)
        {
            J[ct-1] = raw_jump[ct-1] = raw_distortion[ct-1] - raw_distortion[ct];
//            J[ct-1] = D[ct] - D[ct-1];
//            qDebug() << "J: " << J[ct-1];
        }
        ++ct;
    }
        // *** Do not know what the deal is here, dispensing with the exponent, D, and just computing
        //off of raw Mahalanobis distance here (raw_distortion)

    wv::float_t jump_thresh = boost::lexical_cast<wv::float_t>( session_profile.Par("KMeans++", "Jump threshold") ),
            max_jump = wv_stats_max(J, 1, _numClusterings-1);
    int k_idx = (max_jump < jump_thresh)
            ? 0
            : (int)wv_stats_max_index(J, 1, _numClusterings-1);
    Log(distortion_msg);

    if ( ut::Keys(*clusters).at(k_idx)==0)
    {
        Log("ERROR: Unexpected value, KMeansTO::MaxJump.  This does not affect the validity " \
            "of the results but please report this message to mattphillipsphd@gmail.com");
        return 0;
    }
    return ut::Keys(*clusters).at(k_idx) - 1;
        //-1 because the kmfile is 1-based not 0-based

}

//Could also use Kolmogorov-Smirnov, choose k whose worst cluster deviates least from the normal distribution
wv::float_t KMeansTO::Dispersion(const int* assignment, int c)
{
    const wv_matrix* waves = Waves().get();
    int num_records = NumRecords();
    wv::float_t dsum = 0.0;
    for (int i=0; i<num_records; ++i)
    {
        if (assignment[i] != c) continue;
        wv_vector_const_view vi = wv_matrix_const_row(waves, i);
            // *** Need to convert this to subwaves, if this function is ever actually used
        for (int i2=i+1; i2<num_records; ++i2)
        {
            if (assignment[i2] != c) continue;
            wv_vector_const_view vi2 = wv_matrix_const_row(waves, i2);
#ifdef USE_GSL
            wv::float_t d = mgsl::Dist2(&vi.vector, &vi2.vector);
#else
            wv::float_t d = msl::Dist2(&vi.vector, &vi2.vector);
#endif
            dsum += d;
        }
    }
    return dsum;
}

wv::float_t KMeansTO::W(const int* assignment, int k)
{
    //NB, KMeans writes the clusters 0, 1, 2...
    //First get the number of records in each cluster.
    wv::float_t* num_crecs = (wv::float_t*)calloc(k, sizeof(wv::float_t));
    int num_records = NumRecords();
    for (int i=0; i<num_records; ++i)
    {
        int c = assignment[i];
        num_crecs[c] = num_crecs[c] + 1;
    }

    //Now get the dispersions
    wv::float_t w = 0;
    for (int c=0; c<k; ++c)
    {
        wv::float_t d = (1.0/(2.0*num_crecs[c])) * Dispersion(assignment, c);
        w +=d;
    }

    return w;
}

