#include "dunn.h"

Dunn::Dunn(MatCPtr waves, const wv_vector_int* vclusters, const std::vector<int>& clusters, int j_begin, int j_end)
    : _clusters(clusters), _selectedSamplesBegin(j_begin), _selectedSamplesEnd(j_end),
      _vClusters(vclusters), _waves(waves)
{
    _dunnIndexes = std::vector<wv::float_t>(clusters.size());
}

void Dunn::Calculate()
{
    int num_clusters = _clusters.size();
    if (num_clusters==1)
    {
        Calculate1();
        return;
    }

    int num_records = _waves->size1,
            num_samples = _selectedSamplesEnd - _selectedSamplesBegin;

    //Get submatrices, each containing all the waves corresponding to a given clustering
        //Create a map from cluster idxs to 0-based vector index
    std::map<int, int> clus_vec_idxs;
    int ct = 0;
    for (auto it : _clusters)
        clus_vec_idxs[it] = ct++;

        //Count the number of records in each cluster for subsequent submatrix allocation
    std::vector<int> submat_count(num_clusters, 0);
    for (int i=0; i<num_records; ++i)
    {
        int clus = wv_vector_int_get(_vClusters, i),
                vec_idx = clus_vec_idxs[clus];
        submat_count[vec_idx]++;
    }

        //Allocate the submatrices
    std::vector<wv_matrix*> submats(num_clusters);
    for (int c=0; c<num_clusters; ++c)
    {
        int count = submat_count[c];
        submats[c] = wv_matrix_alloc(count, num_samples);
    }

        //Fill up the submatrices
    std::vector<int> record_cts(num_clusters, 0); //The count for each cluster
    for (int i=0; i<num_records; ++i)
    {
        int clus = wv_vector_int_get(_vClusters, i),
                vec_idx = clus_vec_idxs[clus];
        wv_vector_const_view row_view = wv_matrix_const_row(_waves.get(), i),
                sub_row = wv_vector_const_subvector_with_stride(&row_view.vector, _selectedSamplesBegin,
                                                                 row_view.vector.stride, num_samples);
        int record_ct = record_cts[vec_idx]++;
        wv_matrix_set_row(submats[vec_idx], record_ct, &sub_row.vector);
    }
    // ***End acquiring submatrices

    //Calculate mean (center) and sd of clusters
    std::vector<wv_vector*> centers(num_clusters);//, sds(num_clusters);
    for (auto it : _clusters)
    {
        int vec_idx = clus_vec_idxs[it];
        centers[vec_idx] = wv_vector_alloc(num_samples);

        const wv_matrix* submat = submats[vec_idx];
        int record_ct = submat->size1;
        for (int j=0; j<num_samples; ++j)
        {
            wv_vector_const_view col_view = wv_matrix_const_column(submat, j);
            wv::float_t center_j = wv_stats_mean(col_view.vector.data, col_view.vector.stride, record_ct);

            wv_vector_set(centers[vec_idx], j, center_j);
        }
    }

    //Calculate the normalized SD for each cluster *pair*.  This is done by projecting each point onto the line
    //defined by the center of each cluster, and taking this as the distance--this value is then used to define
    //SD.  In other words, SD is relativized for each cluster pair as 'the SD along a dimension'.
    //The math for this comes from http://programmers.stackexchange.com/questions/168572/distance-from-point-to-n-dimensional-line/168577
    std::map<int, std::map<int, wv::float_t> > sd_map;
    for (int c1 = 0; c1<num_clusters; ++c1)
    {
        //Cluster number and the corresponding index, and submatrix
        int clus1 = _clusters.at(c1);
        int vec_idx1 = clus_vec_idxs[clus1];
        const wv_matrix* submat = submats[vec_idx1];
        int record_ct = submat->size1;

        wv_vector* dists = wv_vector_alloc(record_ct); //The projected distance to center1

        for (int c2 = 0; c2<num_clusters; ++c2)
        {
            if (c1==c2) continue;

            int clus2 = _clusters.at(c2);

            //Centers and the vector between them.  We're imagining center1 at the origin, and
            //calculating the SD for center 1
            const wv_vector* center1 = centers[c1],
                    * center2 = centers[c2];
            wv_vector* dists_to_line = wv_vector_alloc(record_ct),
                    * projns_on_line = wv_vector_alloc(record_ct);
#ifdef USE_GSL
                mgsl::PointsToNLine(center1, center2, submat,
                                   projns_on_line, dists_to_line);
#else
                msl::PointsToNLine(center1, center2, submat,
                                  projns_on_line, dists_to_line);
#endif

            wv::float_t sd = wv_stats_sd(projns_on_line->data, projns_on_line->stride, record_ct);
            sd_map[clus1][clus2] = sd;
        }
        wv_vector_free(dists);
    }

    //Now get the modified Dunn index for this cluster.  NB, using the SD as a measure of within-cluster distance
    //is only *one* way to do this--can also use medians/IQR, mean distance between points, etc.
    for (auto it : _clusters) // ***Multi-thread this
    {
        int vec_idx = clus_vec_idxs[it];
        const wv_vector* center = centers[vec_idx];
        int record_ct = submat_count[vec_idx];

        wv::float_t min_dval = std::numeric_limits<wv::float_t>::max();
        for (auto itj : _clusters)
        {
            if (itj == it) continue;
            int vec_idx_j = clus_vec_idxs[itj];
            const wv_vector* center_j = centers[vec_idx_j];
            wv::float_t sd = sd_map.at(it).at(itj),
                    sd_j = sd_map.at(itj).at(it);
            int record_ct_j = submat_count[vec_idx_j];

            //The actual calculation of the index, with SDs *weighted by # of records in the cluster*
            wv_vector* temp_vec = wv_vector_alloc(num_samples);
            wv_vector_memcpy(temp_vec, center_j);
            wv_vector_scale(temp_vec, -1.0);
            wv_vector_add(temp_vec, center);
#ifdef USE_GSL
            wv::float_t dist = gsl_blas_dnrm2(temp_vec),
#else
            wv::float_t dist = msl::blas_nrm2(temp_vec->size, temp_vec->data, temp_vec->stride),
#endif
                    mean_sd = (record_ct*sd + record_ct_j*sd_j) / (record_ct + record_ct_j),
                    dval = dist / mean_sd;
            if (dval < min_dval) min_dval = dval;
            wv_vector_free(temp_vec);
        }

        //Record value
        _dunnIndexes[vec_idx] = min_dval;
    }

    //Free data
    for (int i=0; i<num_clusters; ++i)
    {
        wv_matrix_free( submats[i] );
        wv_vector_free( centers[i] );
    }
}

void Dunn::Calculate1()
{
    int num_records = _waves->size1,
            num_samples = _selectedSamplesEnd - _selectedSamplesBegin;
    wv_vector* centers = wv_vector_alloc(num_samples),
            * vars = wv_vector_alloc(num_samples);

    //Get mean vector, variance
    for (int j=_selectedSamplesBegin, k=0; j<_selectedSamplesEnd; ++j, ++k)
    {
        wv_vector_const_view col_view = wv_matrix_const_column(_waves.get(), j);
        wv::float_t mean_j = wv_stats_mean(col_view.vector.data, col_view.vector.stride, num_records);
        wv::float_t var_j = wv_stats_sd_m(col_view.vector.data,
                                       col_view.vector.stride, num_records, mean_j);

        wv_vector_set(centers, k, mean_j);
        wv_vector_set(vars, k, var_j);
    }

    //Get wave amplitude and the maximum variance.
    wv::float_t min, max;
    wv_stats_minmax(&min, &max, centers->data, centers->stride, num_samples);
    wv::float_t amp = max - min,
            max_var = wv_stats_max(vars->data, vars->stride, num_samples);

    //Record value
    _dunnIndexes[0] = amp / max_var;

    //Free data
    wv_vector_free(centers);
    wv_vector_free(vars);
}
