#include "covthreadmgr.h"

CovThreadMgr::CovThreadMgr(const gsl_matrix* input_mat, gsl_matrix* cov_mat)
    : _inputMat(input_mat), _covMat(cov_mat)
{
    _numThreads = QThread::idealThreadCount();
}
CovThreadMgr::~CovThreadMgr()
{
    for (int i=0; i<_numThreads; ++i)
    {
        _threads[i]->quit();
        delete _covarianceTOs[i];
    }
}

void CovThreadMgr::Calculate()
{
    _numActiveThreads = _numThreads;

    size_t num_rows = _inputMat->size1,
        num_cols = _inputMat->size2;

    for (int i=0; i<_numThreads; ++i)
    {
        QThread* thread = new QThread();
        CovarianceTO* to = new CovarianceTO(this, num_rows);
        to->moveToThread(thread);
        connect(this, SIGNAL(StartThread()), to, SLOT(Start()), Qt::QueuedConnection);
        connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));
        connect(this, SIGNAL(Finished()), to, SLOT(Finished()), Qt::DirectConnection);
            //It's important that this be a direct connection, to avoid a race condition
        connect(to, SIGNAL(ThreadComplete()), this, SLOT(ThreadComplete()), Qt::DirectConnection);
        thread->start();

        emit StartThread();
        disconnect(this, SIGNAL(StartThread()), to, SLOT(Start()));

        _threads.push_back(thread);
        _covarianceTOs.push_back(to);
    }

    _tasksRemaining = num_cols * (num_cols + 1) / 2;
    for (size_t j1 = 0; j1 < num_cols; ++j1)
        for (size_t j2 = j1; j2 < num_cols; ++j2)
        {
            gsl_vector_const_view v1 = gsl_matrix_const_column(_inputMat, j1),
                v2 = gsl_matrix_const_column(_inputMat, j2);

            CovarianceTO::CovSpec cov_spec =
            {
                j1, j2,
                v1.vector.data, v2.vector.data,
                v1.vector.stride, v2.vector.stride
            };
            _covarianceTOs[j2 % _numThreads]->Enqueue(cov_spec);
        }

    while (_numActiveThreads>0)
        SleeperThread::Sleep(25);
}

void CovThreadMgr::SetResult(size_t j1, size_t j2, double covariance)
{
    _mutex.lock();

    gsl_matrix_set(_covMat, j1, j2, covariance);
    gsl_matrix_set(_covMat, j2, j1, covariance);
        //The second statement is redundant if j1==j2, but this is probably
        //faster than testing j1==j2 every time.

    if (--_tasksRemaining == 0) emit Finished();

    _mutex.unlock();
}

void CovThreadMgr::ThreadComplete()
{
    _mutex.lock();
    --_numActiveThreads; //### Need to check thread->IsRunning(), see HistWgtThreadMgr
    _mutex.unlock();
}
