#include "mgsl.h"

//template<typename T>
//wv::GslType<T>::type* mgsl::AllocateRadix2(wv::GslType<T>::type* mat)
gsl_matrix* mgsl::AllocateRadix2(MatCPtr mat)
{
    //First get the new number of samples, minimum 8, below 32 round up, above 32 round to nearest
    int num_rows = mat->size1,
        num_cols_orig = mat->size2,
        num_cols;
    if (num_cols_orig<9)
        num_cols = 8;
    else if (num_cols_orig<48)
    {
        if (num_cols_orig<16)
            num_cols = 16;
        else
            num_cols = 32;
    }
    else
    {
        uint k = 32;
        while (k < (uint)num_cols_orig)
            k <<= 1;
        int k_lower = k>>1,
            k_upper = k;
        num_cols = (num_cols_orig-k_lower < k_upper-num_cols_orig)
                          ? k_lower : k_upper;
    }

    //Initialize new matrix and interpolated sampling vector, and accelerator(?)
    gsl_matrix* transform = gsl_matrix_alloc(num_rows, num_cols);
    double* x_interp = new double[num_cols];
    double scale = (double)(num_cols_orig-1) / (double)(num_cols-1);
    for (int j=0; j<num_cols; ++j)
        x_interp[j] = scale * (double)j;

    Radix2Mgr::Radix2Spec radix2_spec(mat);
    radix2_spec.num_cols = num_cols;
    radix2_spec.x_interp = x_interp;

    Radix2Mgr mgr(radix2_spec, transform);
    mgr.StartThreads();
    mgr.Wait();

    delete[] x_interp;

    return transform;
}

void mgsl::Covariance(const gsl_matrix* mat, gsl_matrix* cov_mat)
{
    size_t num_rows = mat->size1,
        num_cols = mat->size2;

    for (size_t j1 = 0; j1 < num_cols; ++j1)
        for (size_t j2 = j1; j2 < num_cols; ++j2)
        {
            double covariance = gsl_stats_covariance(
                        gsl_matrix_const_column(mat, j1).vector.data,
                        gsl_matrix_const_column(mat, j1).vector.stride,
                        gsl_matrix_const_column(mat, j2).vector.data,
                        gsl_matrix_const_column(mat, j2).vector.stride,
                        num_rows);

            gsl_matrix_set(cov_mat, j1, j2, covariance);
            gsl_matrix_set(cov_mat, j2, j1, covariance);
                //The second statement is redundant if j1==j2, but this is probably
                //faster than testing j1==j2 every time.
        }
}
void mgsl::CovarianceP(const gsl_matrix* mat, gsl_matrix* cov_mat)
{
    CovThreadMgr cov_thread_mgr(mat, cov_mat);
    cov_thread_mgr.StartThreads();
    cov_thread_mgr.Wait();
}
    // ***This actually NOT that much faster than the 1-core version--on a 4-core machine, it
    //produces a 43% reduction in speed.  Has to do with the way things get enstd::queued I think.

#ifdef HAS_CUDA
void mgsl::CovarianceCUDA(const gsl_matrix* mat, gsl_matrix* cov_mat)
{
    size_t num_rows = mat->size1,
        num_cols = mat->size2;

    float* X = new float[num_rows*num_cols],
            * covm = new float[num_cols*num_cols];

    for (size_t i=0; i < num_rows; ++i)
        for (size_t j=0; j < num_cols; ++j)
            X[ IDX2C(i, j, num_rows) ] = gsl_matrix_get(mat, i, j);

    mcu::CovMat(X, num_rows, num_cols, covm);

    for (size_t i=0; i < num_cols; ++i)
        for (size_t j=0; j < num_cols; ++j)
            gsl_matrix_set(cov_mat, i, j, covm[ IDX2C(i, j, num_cols) ]);

    delete X;
    delete covm;
}
void mgsl::CublasSgemm(CBLAS_TRANSPOSE_t cbtA, CBLAS_TRANSPOSE_t cbtB,
                 double alpha, const gsl_matrix* A, const gsl_matrix* B,
                 double beta, gsl_matrix* C)
// ### m, n, k, depend on whether the matrix is transposed
{
    size_t m = A->size1,
           n = B->size2,
           k = A->size2;
    float* Af = new float[m*k],
         * Bf = new float[k*n],
         * Cf = new float[m*n];
    for (size_t i=0; i<m; ++i)
        for (size_t j=0; j<k; ++j)
            Af[ IDX2C(i, j, m) ] = (double)gsl_matrix_get(A, i, j);
    for (size_t i=0; i<k; ++i)
        for (size_t j=0; j<n; ++j)
            Bf[ IDX2C(i, j, k) ] = (double)gsl_matrix_get(B, i, j);

    MCU_MAT_MULT_OP transA, transB;
    switch (cbtA)
    {
        case CblasTrans:
            transA = MCU_TRANSPOSE;
            break;
        case CblasNoTrans:
            transA = MCU_NONE;
            break;
        case CblasConjTrans:
            transA = MCU_CTRANSPOSE;
            break;
    }
    switch (cbtB)
    {
        case CblasTrans:
            transB = MCU_TRANSPOSE;
            break;
        case CblasNoTrans:
            transB = MCU_NONE;
            break;
        case CblasConjTrans:
            transB = MCU_CTRANSPOSE;
            break;
    }

    float alphaf = (double)alpha, betaf = (double)beta;
    mcu::Sgemm( transA, transB,
              m, n, k,
              &alphaf,
              Af, m,
              Bf, k,
              &betaf,
              Cf, m);

    for (size_t i=0; i<m; ++i)
        for (size_t j=0; j<n; ++j)
            gsl_matrix_set(C, i, j, Cf[ IDX2C(i, j, m) ] );

    delete Af;
    delete Bf;
    delete Cf;
}
#endif

void mgsl::PCATransform(const gsl_matrix* mat, gsl_matrix* trans_mat,
                        gsl_matrix* eigenvecs, gsl_vector* eigenvals, bool use_cuda)
{
    int size1 = mat->size1,
            size2 = mat->size2;

    //Generate the output matrix which is the input to gsl_eigen_symmv.  This gets destroyed
    //in the process (!) so apparently can't be the same as the output matrix
//    qDebug() << "PCATransform 1, " << size1 << ", " << size2;
    gsl_matrix* input = gsl_matrix_alloc(size1,size2);
//    qDebug() << "PCATransform 2";
    gsl_matrix_memcpy(input, mat);

    //These hold the resultant eigenvectors and eigenvalues.  If the user wants them he must allocate them
    //in the calling function
    bool free_eigenvecs = !eigenvecs,
            free_eigenvals = !eigenvals;
//    qDebug() << "PCATransform 3";
    if (!eigenvecs) eigenvecs = gsl_matrix_alloc(size2,size2);
    if (!eigenvals) eigenvals = gsl_vector_alloc(size2);

//    qDebug() << "PCATransform 4";
    PrinComp(input, eigenvecs, eigenvals, use_cuda);

    //Now the adjusted data
    gsl_matrix_view ev = gsl_matrix_submatrix(eigenvecs, 0, 0, size2, trans_mat->size2);
#ifdef HAS_CUDA
    if (use_cuda)
        CublasSgemm(CblasNoTrans, CblasNoTrans, 1.0, mat, &ev.matrix, 0.0, trans_mat);
    else
#endif
        gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, mat, &ev.matrix, 0.0, trans_mat);

    gsl_matrix_free(input);
    if (free_eigenvecs) gsl_matrix_free(eigenvecs);
    if (free_eigenvals) gsl_vector_free(eigenvals);
}

void mgsl::PointsToNLine(const gsl_vector* line_pt1, const gsl_vector* line_pt2, const gsl_matrix* pts,
                        gsl_vector* projns_on_line, gsl_vector* dists_to_line)
{
    //Temporary vectors
    int num_dims = line_pt1->size;
    gsl_vector* dline = gsl_vector_alloc(num_dims), //Distance between line points
            * dpoint = gsl_vector_alloc(num_dims), //Distance of pt to line_pt1
            * line_pt3 = gsl_vector_alloc(num_dims); //for storing t*dcenter

    //Get vector between line points, line_pt2 - line_pt1
    gsl_vector_memcpy(dline, line_pt1);
    gsl_vector_scale(dline, -1.0);
    gsl_vector_add(dline, line_pt2);

    int num_pts = pts->size1;
    for (int i=0; i<num_pts; ++i)
    {
        gsl_vector_const_view data_ptv = gsl_matrix_const_row(pts, i);

        //Get distance from pt to line_pt1
        gsl_vector_memcpy(dpoint, line_pt1);
        gsl_vector_scale(dpoint, -1.0);
        gsl_vector_add(dpoint, &data_ptv.vector);

        //Do the math--use dot product etc. to get distance to line
        double dot_point, dot_line, t;
        dot_point = cblas_ddot(dpoint->size, dpoint->data, dpoint->stride, dline->data, dline->stride);
            //dot_point increases with similarity of dline to dpoint--that's what the dot product does
        dot_line = cblas_dnrm2(dline->size, dline->data, dline->stride);
        dot_line *= dot_line;
            //dot_line is the squared distance between the two line points, equivalent to the dot product
            //of dline with itself

        t = dot_point/dot_line;
            //t in [0, 1] measures how close the dpoint vector is to falling on the line

        gsl_vector_memcpy(line_pt3, dline);
        gsl_vector_scale(line_pt3, t);
            //line_pt3 is now a vector corresponding to dline, magically scaled down by t so that the
            //vector [line_pt3, data_pt] is orthogonal to the line
        gsl_vector_sub(dpoint, line_pt3);
            //Now, dpoint has been translated so that, effectively, its origin is now line_pt3.  It is the
            //vector from line_pt3 to data_pt and so, the crow-flies vector of data_pt to the line.

        double dist_to_line = cblas_dnrm2(dpoint->size, dpoint->data, dpoint->stride);
            //The length of the [line_pt3, data_pt] vector which is the distance to the line.
        gsl_vector_set(dists_to_line, i, dist_to_line);

        double dist_on_line = cblas_dnrm2(line_pt3->size, line_pt3->data, line_pt3->stride);
        gsl_vector_set(projns_on_line, i, (t>0) ? dist_on_line : -dist_on_line);
    }
    gsl_vector_free(dline);
    gsl_vector_free(dpoint);
    gsl_vector_free(line_pt3);
}

void mgsl::PrinComp(const gsl_matrix* mat,
                    gsl_matrix* eigenvecs, gsl_vector* eigenvals, bool use_cuda)
{
    int M = mat->size1, //Number of data entries (rows)
        N = mat->size2; //Number of points per data entry (columns)

//    qDebug() << "Princomp 1, " << M << ", " << N;
    //Get the covariance matrix;
    gsl_matrix* cov_mat = gsl_matrix_alloc(N, N);
#ifdef HAS_CUDA
    if(use_cuda)
        mgsl::CovarianceCUDA(mat, cov_mat);
    else
#endif
    {
        //de-mean the columns
        gsl_matrix* mat_m = gsl_matrix_alloc(M, N);
        gsl_matrix_memcpy(mat_m, mat);
        for (int j=0; j<N; ++j)
        {
            gsl_vector_view v = gsl_matrix_column(mat_m, j);
            double m = gsl_stats_mean(v.vector.data, v.vector.stride, M);
            gsl_vector_add_constant(&v.vector, -m);
        }

        //Covariance now is just the following matrix multiplication followed
        //by the scaling
        const double recip = 1.0/(double)(M>1 ? M-1 : 1);
        gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, mat_m, mat_m, 0.0, cov_mat);
        gsl_vector vec;
        vec.size = N*N;
        vec.stride = 1;
        vec.block = cov_mat->block;
        vec.data = cov_mat->data;
        vec.owner = 0;
        gsl_blas_dscal(recip, &vec);
        gsl_matrix_free(mat_m);
    }

//    qDebug() << "Princomp 2, ";
    //Now get the eigenvectors of the covariance matrix, which are the
    //principal components
    gsl_eigen_symmv_workspace* ws = gsl_eigen_symmv_alloc(N);
//    qDebug() << "Princomp 3, ";
    gsl_eigen_symmv(cov_mat, eigenvals, eigenvecs, ws);
    gsl_matrix_free(cov_mat);
    gsl_eigen_symmv_free(ws);
//    qDebug() << "Princomp 4, ";
}

double mgsl::Dist2(const gsl_vector* v1, const gsl_vector* v2)
{
    int length = v1->size;
    double dsum;
    gsl_vector* temp = gsl_vector_alloc(length);
    gsl_vector_memcpy(temp, v1);
    gsl_vector_sub(temp, v2);
    dsum = cblas_dnrm2(length, temp->data, temp->stride);
    return dsum * dsum;
}
