#include "clusterto.h"
#include "../../../memrep/specs/clusterspec.h"
#include "../../../gui/mainwindow.h"

ClusterTO::ClusterTO(int thd_num, const std::vector<wv::CLUSTER_ID>& cids)
    : WorkerTO(thd_num), _cids(cids)
{
}

void ClusterTO::Start()
{
   for (wv::CLUSTER_ID cid : _cids)
   {
        try
        {
            Cluster(cid);
        }
        catch (WException& e)
        {
            Log("WARNING: Clustering error on ID " + std::to_string(cid)
                + ": " + e.what() + ".  This clustering aborted");
        }
            //So, we do *not* break if a single file throws.
   }

    WorkFinished();
}

void ClusterTO::Cluster(wv::CLUSTER_ID cid)
{
    ClusterMediator& cmed = ClusterMediator::Instance();
    CluSpecPtr clu_spec = cmed.GetCluSpecPtr(cid);
    short electrode = cmed.GetElectrode(cid);
    InputFilePtr input_file = cmed.GetInputFile(cid);
    WaveData* wave_data = input_file->WData(electrode);

    wv::TRANSFORM transform = clu_spec->Transform();
    int tid;
    if (transform==wv::USER)
    {
        //What we have to do in this case is *build* the user transform for this session/electrode
        //out of the elements from the transform specified in WaveSpaceGui when batch clustering starts.
        InputFilePtr old_input_file = cmed.GetInputFile(wv::DEFAULT_CID);
        CluSpecPtr gui_clu_spec = cmed.GetCluSpecPtr(wv::DEFAULT_CID);
        WSGSpec& gui_wsg_spec = gui_clu_spec->GetWSGSpec();
        int gui_tid = gui_wsg_spec.TransformId( gui_wsg_spec.ActiveAxis() );

        if (wave_data->HaveTransform(transform, gui_tid))
            tid = gui_tid;
            //In this case, we are not in batch mode, so there is no need to create a new transform
        else
        {
            short gui_electrode = cmed.GetElectrode(wv::DEFAULT_CID);
            const UTS& gui_spec = old_input_file->WData(gui_electrode)->UTSpec(gui_tid);
            tid = wave_data->CreateUserTransform(gui_spec);
            clu_spec->SetTransformId(tid);
        }
    }
    else
        tid = wv::DEFAULT_TID;
    wv::CLUSTER_METHOD method = clu_spec->ClusterMethod();

    MatCPtr waveforms = wave_data->Waveforms(transform, tid);
    if (!waveforms)
        throw EmptyTransform("ClusterTO::Cluster", "No waveforms available for this selection");

    // ****Start running the analysis, using current parameters. ****
        //Get filename
    std::string wfm_file = FileBase::MakeTempFileName(FileBase::UserPath(), input_file->Name(), transform,
                                             tid, "wfm", electrode);

        //Create cluster method
    ClusMethodTOBase* cluster_method; // ### Memory leak, as AnalysisTOBase holds a MatCPtr
    if (method==wv::MANUAL)
        cluster_method = ClusMethodTOBase::Create(cid, method, wave_data, wfm_file);
    else
        cluster_method = ClusMethodTOBase::Create(cid, transform, method, waveforms, wfm_file);
    cluster_method->SetWindow(0, waveforms->size1);

    //Set up the processing thread--Do NOT know why this extra thread is necessary
    QThread* thread = new QThread();
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));
    cluster_method->moveToThread(thread);
    connect(cluster_method, SIGNAL(Finished(void*)), cmed.GetMainWindow(), SLOT(ClusterFinished(void*)), Qt::QueuedConnection);

    connect(this, SIGNAL(StartClustering()), cluster_method, SLOT(Start()), Qt::QueuedConnection);
    thread->start();
    emit StartClustering();
    disconnect(this, SIGNAL(StartClustering()), cluster_method, SLOT(Start()));
    // ***Very strange--just calling ClusterMethodTOBase::Start directly, as in below, does NOT work.  Have to create
    //this redundant extra thread.

//    cluster_method->Start();
}
