#include "histscatterto.h"

HistScatterTO::HistScatterTO(int thd_num, const HistWidget* hist_widget, HistSpec hist_spec,
                             wv_vector* xdata_vec, wv_vector* ydata_vec) :
    WorkerTO(thd_num), _histWidget(hist_widget), _scaleSpec(hist_widget->_scaleSpec),
    _histSpec(hist_spec),
    _xDataVec(xdata_vec), _yDataVec(ydata_vec)
{
}
HistScatterTO::~HistScatterTO()
{
    wv_vector_free(_xDataVec);
    wv_vector_free(_yDataVec);
}

void HistScatterTO::Start()
{
    //Get information from _histSpec, mostly for convenience.
    int num_bins_x = _histSpec.num_bins_x,
        num_bins_y = _histSpec.num_bins_y,
        start_row = _histSpec.start_row,
        end_row = _histSpec.end_row;
    const boost::dynamic_bitset<>& is_accepted = _histSpec.is_accepted;
    const WaveSpaceGuiSpec& spec = _histSpec.spec;
    int range = end_row - start_row;

    //Create subvectors
    wv_vector_view xdata_view = wv_vector_subvector_with_stride(_xDataVec, start_row, _xDataVec->stride, range),
            ydata_view = wv_vector_subvector_with_stride(_yDataVec, start_row, _yDataVec->stride, range);
    wv_vector* xzoom = &xdata_view.vector,
            * yzoom = &ydata_view.vector;

    //Scale and zoom
    _scaleSpec->ValToPixXVec(xzoom),
    _scaleSpec->ValToPixYVec(yzoom);

    //Allocate and initialize matrices
    wv_matrix_int* mat_accept = wv_matrix_int_alloc(num_bins_x, num_bins_y),
                * mat_unaccept = wv_matrix_int_alloc(num_bins_x, num_bins_y);
    wv_matrix_int_set_zero(mat_accept);
    wv_matrix_int_set_zero(mat_unaccept);

    bool accept_nonempty = !is_accepted.empty(),
            use_transform = spec.HistLimitMode()==WSGSpec::TRANSFORM;
    wv::float_t time_begin = spec.TimeBegin(),
            time_end = spec.TimeEnd();
    if (use_transform)
        for (int row=0; row<range; ++row)
        {
            int i = wv_vector_get(xzoom, row),
                    j = wv_vector_get(yzoom, row);
            if (i<0 || j<0 || i>=num_bins_x || j>=num_bins_y) continue;
            if (accept_nonempty && is_accepted.test(row+start_row))
                wv_matrix_int_set(mat_accept, i, j,
                                   wv_matrix_int_get(mat_accept, i, j)+1 );
            else
                wv_matrix_int_set(mat_unaccept, i, j,
                                   wv_matrix_int_get(mat_unaccept, i, j)+1 );
        }
    else
        for (int row=0, orow=start_row; row<range; ++row, ++orow)
        {
            int i = wv_vector_get(xzoom, row),
                    j = wv_vector_get(yzoom, row);
            if (i<0 || j<0 || i>=num_bins_x || j>=num_bins_y) continue;
            wv::float_t timestamp = spec.Timestamp(orow);
            if (timestamp >= time_begin && timestamp <= time_end)
            {
                if (accept_nonempty && is_accepted.test(orow))
                    wv_matrix_int_set(mat_accept, i, j,
                                       wv_matrix_int_get(mat_accept, i, j)+1 );
                else
                    wv_matrix_int_set(mat_unaccept, i, j,
                                       wv_matrix_int_get(mat_unaccept, i, j)+1 );
            }
                //Get it--if the record falls outside the time window, it isn't included in the histogram.
        }
    //A little inelegant to repeat the loop, but the speed gain is probably non-negligible

    auto result = new std::pair<wv_matrix_int*, wv_matrix_int*>(mat_accept, mat_unaccept);
    WorkFinished(result);
}
