#include "transformtobase.h"
#include "../../file/output/wfmfileout.h"
#ifdef USE_GSL
#include "fourierto.h"
#include "splineto.h"
#endif
#include "haarto.h"
#include "pca_to.h"

TransformTOBase::TransformTOBase(MatCPtr waves, const std::string& input_file_path, short electrode,
                                 ParamsBase* pars, bool is_raw)
    : AnalysisTOBase(waves, waves->size1, input_file_path), _electrode(electrode), _isRaw(is_raw)
{
    SetParams( ParamsCPtr(pars) );
}
TransformTOBase::~TransformTOBase()
{
    _mat.reset(); //Deleted != out of scope, so you actually need to do this apparently
}
TransformTOBase* TransformTOBase::Create(wv::TRANSFORM transform, MatCPtr waves,
#ifdef USE_GSL
                                         wv::float_t us_per_samp,
#endif
                                         const std::string& input_file,
                                        short electrode)
{
    TransformTOBase* transform_to;
    switch (transform)
    {
#ifdef USE_GSL
        case wv::FOURIER:
            transform_to = new FourierTO(waves, input_file, electrode, NULL, true);
            break;
        case wv::FOURIER_D:
            transform_to = new FourierTO(waves, input_file, electrode, NULL, false);
            break;
        case wv::SPLINE:
            transform_to = new SplineTO(waves, us_per_samp, input_file, electrode, NULL, true);
            break;
        case wv::SPLINE_D:
            transform_to = new SplineTO(waves, us_per_samp, input_file, electrode, NULL, false);
            break;
#endif
        case wv::WAVELET:
            transform_to = new HaarTO(waves, input_file, electrode, NULL, true);
            break;
        case wv::WAVELET_D:
            transform_to = new HaarTO(waves, input_file, electrode, NULL, false);
            break;
        case wv::PCA:
            transform_to = new PCA_TO(waves, input_file, electrode, NULL, true);
            break;
        case wv::PCA_D:
            transform_to = new PCA_TO(waves, input_file, electrode, NULL, false);
            break;
        default:
            throw BadTransform("TransformTOBase::Create", wv::TransformName(transform));
    }
    return transform_to;
}
void TransformTOBase::AnalysisFinished(void* data)
{
    auto pair = static_cast< std::pair<wv::TRANSFORM,MatPtr>* >(data);
    _mat = pair->second;
    WriteTempFile(_mat);
    AnalysisTOBase::AnalysisFinished(pair);
}
void TransformTOBase::AnalysisFinished(const void* data) const
{
    auto pair = static_cast< const std::pair<wv::TRANSFORM, const wv_matrix*>* >(data);
    WriteTempFile( MatCPtr(pair->second) );
    AnalysisTOBase::AnalysisFinished(pair);
}

void TransformTOBase::WriteTempFile(MatCPtr transform, wv::TRANSFORM tform) const
{
    if (tform==wv::UNKNOWN_TRANSFORM) tform = Transform();
    std::string full_path = FileBase::MakeTempFileName(FileBase::UserPath(), InputFileName(), tform,
                                              wv::DEFAULT_TID, "wfm", _electrode);
        //Since transforms aren't tied to a particular analysis, there is a generic id for all of them

    int num_records = transform->size1,
        num_samples = transform->size2;

    FileBase::WfmHeader header =
    {
        (uchar) 0,
        num_records,
        num_samples,
        sizeof(wv::float_t),
        0,
        1.0,
        0.0
    };

    WfmFileOut temp_file(full_path, transform, header);

    temp_file.Save();
    QApplication::processEvents();
}

std::string TransformTOBase::InputFileName() const
{
    int pos1 = InputFullPath().find_last_of('/') + 1;
    std::string file_name = InputFullPath().substr(pos1);
    return file_name;
}
