#include "waveformsinbase.h"
#include "apmfile.h"
#include "asciifilein.h"
#include "fetfile.h"
#include "mexfile.h"
#include "nevfile.h"
#include "plxfile.h"

const int WaveformsInBase::UTAH_ARRAY_96 = 96;

WaveformsInBase::WaveformsInBase(const std::string& full_path, bool is_binary)
    : InputFileBase(full_path, is_binary), _progressBar(new QProgressBar), _isBatch(false), _usPerSample(1.0)
{
}
WaveformsInBase::~WaveformsInBase()
{
    std::map<short, WaveData*>* mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    if (mwdata)
    {
        for (auto it : *mwdata)
            delete it.second;
        delete mwdata;
    }
}

WaveformsInBase* WaveformsInBase::Create(const std::string& full_path)
{
    WaveformsInBase* waves_file(NULL);
    if ( ut::EndsWith(full_path, ".apm") )
        waves_file = new APMFile(full_path);
    else if ( ut::EndsWith(full_path, ".plx") )
        waves_file = new PlxFile(full_path);
    else if ( ut::EndsWith(full_path, 'M') )
        waves_file = new MexFile(full_path);
    else if ( ut::Contains(full_path, ".fet.") )
        waves_file = new FetFile(full_path);
    else if ( ut::EndsWith(full_path, ".txt") )
        waves_file = new AsciiFileIn(full_path);
    else if ( ut::EndsWith(full_path, ".nev") )
        waves_file = new NevFile(full_path);
    if (!waves_file)
        throw BadFileName("WaveformsInBase::Create", full_path);
    return waves_file;
}

void WaveformsInBase::FinishedLoading()
{
    auto mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    for (auto it : *mwdata)
        it.second->DeleteFPs();
}
int WaveformsInBase::GenerateTransforms(short elc)
{
    //Generate the transforms etc. for each WaveData object
    int num_records_used = 0;
    auto mwdata = static_cast< std::map<short, WaveData*>* >( Data() );
    if (elc==-1)
        for (auto it : *mwdata)
        {
            num_records_used += it.second->Preprocess();
            it.second->Unload();
        }
    else
        num_records_used = (*mwdata)[elc]->Preprocess();
    return num_records_used;
}
void WaveformsInBase::MakeWaveData(const std::vector<short>& v_electrodes, bool has_online_clus)
{
    //Get the electrodes, and determine the number of records in each one.
    std::vector<short> electrode_set = VElectrodes();
    int num_electrodes = electrode_set.size();
    _numWaves = std::vector<int>(num_electrodes, 0);
    for (auto elc : v_electrodes)
        for (int j=0; j<num_electrodes; ++j)
        {
            short elc_j = electrode_set.at(j);
            if (elc_j==elc)
            {
                _numWaves[j]++;
                break;
            }
        }

    //Create the waveform files and write in the header.  Do *not* allocate any memory here.
    std::map<short, WaveData*>* mwdata = new std::map<short, WaveData*>();
    for (int i=0; i<num_electrodes; ++i)
    {
        short elc = electrode_set.at(i);
        WaveData* wave_data = new WaveData(this, elc, _usPerSample );
        wave_data->SetHasOnlineClus(has_online_clus);
//        qDebug() << elc << ", " << _numWaves.at(i);
        wave_data->AllocateMemory(_numWaves.at(i)); //Only does this for the online clustering
        wave_data->SetupWavefiles(_numWaves.at(i), NumSamples()); //For timestamps and raw waveforms
        (*mwdata)[elc] = wave_data;
    }

    SetData(mwdata);
}
