#include "rexefile.h"

const int RexEFile::EHEADER_SZ = 512;
const int RexEFile::REX_PANIC = -123;

RexEFile::RexEFile(const std::string& old_efile, const std::string& name, const WaveData* data, const boost::dynamic_bitset<>& accepted)
    : OutputFileBase(name, true, data, accepted), _EFileName(old_efile),  _EFile(NULL)
{
    if (_EFileName != "") _EFile = fopen(_EFileName.c_str(), "rb");
}
void RexEFile::Save() const
{
    if (_EFile==(FILE*)NULL)
    {
        Log("ERROR: Rex E-File does not exist");
        return;
    }

    int num_records = Data()->NumRecords();
    MatCPtr tmat = const_cast<WaveData*>(Data())->Timestamps(); // ### Oh well...
    wv_vector_const_view tview = wv_matrix_const_column(tmat.get(), 0);
    const wv_vector* timestamps = &tview.vector;
    int code = 610 + 3*Data()->Electrode();
    std::vector<RexEvent> events;
    events.reserve( 4*num_records ); //This should be more than enough

    /****
      So the strategy here is simple and twofold:
      1. Create REX events for each waveform, and place them, along with the pre-existing REX events,
        into a single data structure, with the correct chronological order for all events.
      2. Read the contents of this structure directly into the (new) E file.
      ****/

    //Get header of original E-File, gets rewritten to the new one below.  The header just contains
    //the file name and the date it was collected.
    char header[EHEADER_SZ];
    int br = fread(header, EHEADER_SZ, 1, _EFile);
    if (br != 1)
    {
        Log("ERROR: " + _EFileName + " is corrupted, no waveforms saved.");
        return;
    }
    Log("E-File header: " + std::string(&header[2]));
    QApplication::processEvents();

#ifdef REX_ERRLOG
    oldlogf = fopen("RexOldEvents.log", "w");
#endif

    //Get first REX event
    RexEvent ev = ReadRexEvent();

    //First all of the waveforms occurring before the first REX event
    uint rex_start = ev.tim;
    int wave_idx = 0, //The of the record currently under consideration--may or may not be accepted.
        all_ct = 0; //The count on how many events there are total to be read back into the E file
    while (wave_idx<num_records && wv_vector_get(timestamps, wave_idx)<rex_start)
    {
        if (IsAccepted(wave_idx))
        {
            RexEvent mex_ev(all_ct++, code, wv_vector_get(timestamps, wave_idx));
            events.push_back(mex_ev);
        }
        ++wave_idx;
    }

    //Now waveforms occuring during the session, i.e. intermixed with REX events
    int old_spike_ct = 0, //We display what the change in spike count between pre- and post-sorting is
        new_spike_ct = 0, //NB, only spikes occurring *during* the session count for this, of course.
        weird_code_ct = 0; //E.g., -112 which has out-of-order timestamps.
    while (ev.code != REX_PANIC)
    {
        if (ev.code<600 || ev.code>=800) //Evidently codes 700-800 are reserved for offline MEX spike events.
            //Could not figure out what they were, searching REX or MEX. These are skipped.
            //600-700 are REX spike events, and we DELETE THESE TOO.  Since we're creating
            //a new E-file, there's no point in keeping them.
        {
            if (ev.code<0) ++weird_code_ct;
            //In this loop, we push all records occuring prior to ev, the current *REX event* into the events vector.
            while (wave_idx<num_records && wv_vector_get(timestamps, wave_idx)<ev.tim && ev.code>=0)
            {
                if (IsAccepted(wave_idx))
                {
                    RexEvent mex_ev(all_ct++, code, wv_vector_get(timestamps, wave_idx));
                        // *** The seqNum is -1 in original REX spike events
                    events.push_back(mex_ev);
                    ++new_spike_ct;
                }
                ++wave_idx;
            }
            ev.seq_num = all_ct++; //Replacing the old number
            events.push_back(ev);
        }
        else //if (ev.code>=600 && ev.code<700) // *Lots* of 700 events
            ++old_spike_ct;
        ev = ReadRexEvent();
    }

    Log("WARNING: " + std::to_string(weird_code_ct)
        + " events from old E-File with code<0, timestamps ignored.");

    //Now the waveforms that occurred after the session
    while (wave_idx<num_records)
    {
        if (IsAccepted(wave_idx))
        {
            RexEvent mex_ev(all_ct++, code, wv_vector_get(timestamps, wave_idx));
            events.push_back(mex_ev);
        }
        ++wave_idx;
    }

    // ***Now finally write everything in events to disk.
#ifdef REX_ERRLOG
    FILE* newlogf = fopen("RexNewEvents.log", "w");
    for (auto it=events.begin(); it!=events.end(); ++it)
        fprintf(newlogf, "%hd\t%hd\t%u\n", it->seq_num, it->code, it->tim);
    fclose(newlogf);
    fclose(oldlogf);
#endif

    FILE* new_efile = File();
    fwrite(header, EHEADER_SZ, 1, new_efile);
    int num_events = fwrite(events.data(), sizeof(RexEvent), events.size(), new_efile);
    Log( std::to_string(num_events) + " REX events written to " + Name() + ". " + std::to_string(old_spike_ct)
         + " spikes before sorting, " + std::to_string(new_spike_ct) + " spikes after sorting, "
         "new total " + std::to_string((int)((100.0*new_spike_ct)/old_spike_ct + 0.5) )
         + "% of the old." );
}

//After the header, it's nothing but RexEvents in the E file
RexEFile::RexEvent RexEFile::ReadRexEvent() const
{
    RexEvent ev;
    int num_bytes = fread(&ev, sizeof(RexEvent), 1, _EFile);
    if (num_bytes != 1)
        ev = RexEvent(0,REX_PANIC,0); //The 'panic' value used to be 0, but this came up naturally among the
            //event codes, causing a false positive
#ifdef REX_ERRLOG
    fprintf(oldlogf, "%hd\t%hd\t%u\n", ev.seq_num, ev.code, ev.tim);
#endif
    return ev;
}

