#ifndef WEXCEPTION_H
#define WEXCEPTION_H

#include "wv.h"
#include "ut_def.h"

class WException : public std::exception
{
    public:
        explicit WException(const std::string& func, const std::string& msg, const std::string& name);
        virtual ~WException() throw() {}

        virtual const char* what() const throw() = 0;

        const std::string& Function() const { return _function; }
        const std::string& Message() const { return _message; }
        const std::string& Name() const { return _name; }

    private:
        const std::string _function, _message, _name;
};

class BatchException : public WException
{
    public:
        explicit BatchException(const std::string& func, const std::string& msg)
        : WException(func, msg, "Batch Exception")
        {
        }
        virtual ~BatchException() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadData : public WException
{
    public:
        explicit BadData(const std::string& func, int electrode)
        : WException(func, std::to_string(electrode), "Bad Data"), _electrode(std::to_string(electrode))
        {
        }
        virtual ~BadData() throw() {}

        virtual const char* what() const throw();
    private:
        std::string _electrode;
};

class BadParameter : public WException
{
    public:
        explicit BadParameter(const std::string& func, const std::string& param)
        : WException(func, param, "Bad Parameter")
        {
        }
        virtual ~BadParameter() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class MutexFail : public WException
{
    public:
        explicit MutexFail(const std::string& func)
        : WException(func, "", "Mutex lock failure:")
        {
        }
        virtual ~MutexFail() throw() {}

        virtual const char* what() const throw();
};

class NullTransform : public WException
{
    public:
        explicit NullTransform(const std::string& func)
        : WException(func, "", "Null Transform")
        {
        }
        virtual ~NullTransform() throw() {}

        virtual const char* what() const throw();
};

class BadUserTransform : public WException
{
    public:
        explicit BadUserTransform(const std::string& func, int tid)
        : WException(func, std::to_string(tid), "Bad User Transform")
        {
        }
        virtual ~BadUserTransform() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadClusterMethod : public WException
{
    public:
        explicit BadClusterMethod(const std::string& func, wv::CLUSTER_METHOD method)
        : WException(func, wv::ClusterName(method), "Bad Cluster Method") //In general, this will be wv::MANUAL
        {
        }
        virtual ~BadClusterMethod() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadCID : public WException
{
    public:
        explicit BadCID(const std::string& func, wv::CLUSTER_ID cid)
        : WException(func, std::to_string(cid), "Cluster ID doesn't exist")
        {
        }
        virtual ~BadCID() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadTID : public WException
{
    public:
        explicit BadTID(const std::string& func, wv::TRANSFORM_ID tid)
        : WException(func, std::to_string(tid), "Transform ID doesn't exist or is not stored")
        {
        }
        virtual ~BadTID() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadTransform : public WException
{
    public:
        explicit BadTransform(const std::string& func, const std::string& transform)
        : WException(func, transform, "Bad Transform")
        {
        }
        virtual ~BadTransform() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadWaveGui : public WException
{
    public:
        explicit BadWaveGui(const std::string& func, void* wave_gui)
        : WException(func,
                     std::to_string((unsigned long long)wave_gui),
                     "WaveGui doesn't exist")
        {
        }
        virtual ~BadWaveGui() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadGuiOp : public WException
{
    public:
        explicit BadGuiOp(const std::string& func, const std::string& msg)
        : WException(func, msg, "Bad Gui Operation")
        {
        }
        virtual ~BadGuiOp() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadDrawLine : public WException
{
    public:
        explicit BadDrawLine(const std::string& func, const std::string& msg)
        : WException(func, msg, "Bad Drawing line")
        {
        }
        virtual ~BadDrawLine() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class EmptyTransform : public WException
{
    public:
        explicit EmptyTransform(const std::string& func, const std::string& msg)
        : WException(func, msg, "Empty Transform")
        {
        }
        virtual ~EmptyTransform() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadValue : public WException
{
    public:
        explicit BadValue(const std::string& func, const std::string& msg)
        : WException(func, msg, "Bad Value")
        {
        }
        virtual ~BadValue() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class NullObject : public WException
{
    public:
        explicit NullObject(const std::string& func, const std::string& msg)
        : WException(func, msg, "Null Object")
        {
        }
        virtual ~NullObject() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class WIOException : public WException
{
    public:
    explicit WIOException(const std::string& func, const std::string& file_name, const std::string& msg = "")
        : WException(func, file_name + ": " + msg, "IO Exception")
        {
        }
        virtual ~WIOException() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadFileName : public WIOException
{
    public:
        explicit BadFileName(const std::string& func, const std::string& file_name)
        : WIOException(func, file_name, "Bad File Name")
        {
        }
        virtual ~BadFileName() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

class BadClustering : public WIOException
{
    public:
        explicit BadClustering(const std::string& func, wv::CLUSTER_ID cid)
        : WIOException(func, std::to_string(cid), "Bad Clustering")
        {
        }
        virtual ~BadClustering() throw() {}

        virtual const char* what() const throw() { return WException::what(); }
};

#endif // WEXCEPTION_H
