#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "../globals/globals.h"
#include "../file/input/inputfilebase.h"
#include "../file/input/profilein.h"
#include "../file/input/userclufilein.h"
#include "../file/output/clufileout.h"
#include "../file/output/usertclufileout.h"
#include "../file/output/userclufileout.h"
#include "../file/output/rexefile.h"
#include "../file/output/simplebinary.h"
#include "../file/output/simpleascii.h"
#include "../file/output/userdatafileout.h"
#include "../file/output/parfileout.h"
#include "../analysis/cluster/clusmethodtobase.h"
#include "../analysis/cluster/manualto.h"
#include "../globals/log.h"

#include "../memrep/specs/clusterspec.h"
#include "../analysis/batch.h"
#include "../analysis/channelbatch.h"

#ifndef USE_GSL
#include "../license/licensemgr.h"
#endif

#include "wavespacegui.h"
#include "profilegui.h"
#include "batchgui.h"
#include "mdi/loggui.h"
#include "mdi/paramgui.h"
#include "mdi/electrodegui.h"
#include "mdi/wavegui.h"
#include "mdi/progressgui.h"
#include "widgets/aboutgui.h"
#include "gui.h"

#ifdef HAVE_EXTRA
#include "widgets/wave/gridwavewidget.h"
#include "wavegridgui.h"
#endif

namespace Ui
{
    class MainWindow;
}

class MainWindow : public QMainWindow
{
    Q_OBJECT

    friend class Batch;

    public:
        explicit MainWindow(QWidget *parent = 0);
        ~MainWindow();

#ifdef HAVE_EXTRA
        void AddToWaveGrid(wv::CLUSTER_ID cid);
#endif
        void CreateWaveGui(wv::CLUSTER_ID cid);
        void LaunchWaveGui(wv::CLUSTER_ID cid);

    protected:
        void showEvent(QShowEvent*);
        void closeEvent(QCloseEvent*);

    signals:
        void StartClustering();
        void StopMovie();

    private slots:
        void ReceiveProfile(const std::string& profile);

        void on_actionREX_E_file_triggered() { gui::Slot(this, &MainWindow::OnActionREXEFileTriggered); }
        void on_actionClu_file_triggered() { gui::Slot(this, &MainWindow::OnActionCluFileTriggered); }
        void on_actionTClu_file_triggered() { gui::Slot(this, &MainWindow::OnActionTCluFileTriggered); }
        void on_actionSave_All_triggered() { gui::Slot(this, &MainWindow::OnActionSaveAllTriggered); }
        void on_actionSimple_binary_triggered() { gui::Slot(this, &MainWindow::OnActionSimpleBinaryTriggered); }
        void on_actionSimple_ASCII_triggered() { gui::Slot(this, &MainWindow::OnActionSimpleASCIITriggered); }
        void on_actionLoad_Classification_triggered() { gui::Slot(this, &MainWindow::OnActionLoadClassificationTriggered); }
        void on_actionRunBatch_triggered() { gui::Slot(this, &MainWindow::OnActionRunBatchTriggered); }
        void on_actionLoad_Waveform_File_triggered() { gui::Slot(this, &MainWindow::OnActionLoadWaveformFileTriggered); }

        void on_actionSession_triggered() { gui::Slot(this, &MainWindow::OnActionSessionTriggered); }
        void on_actionKMeans_triggered() { gui::Slot(this, &MainWindow::OnActionKMeansTriggered); }
        void on_actionAutoClass_triggered() { gui::Slot(this, &MainWindow::OnActionAutoClassTriggered); }
        void on_actionSPC_triggered() { gui::Slot(this, &MainWindow::OnActionSPCTriggered); }
        void on_actionKlustaKwik_triggered() { gui::Slot(this,  &MainWindow::OnActionKlustaKwikTriggered); }
        void DisplayParameters(ParamsBase* pars)
        {
            gui::Slot2(this, std::bind(&MainWindow::OnDisplayParameters, this, pars, std::placeholders::_1) );
        }

        void on_actionAbout_triggered();
        void on_actionWaveSorter_Pro_triggered();

    public slots:
        void BatchStart();
        void BatchComplete();
        void ClusterFinished(void* pcid);
            //A signal is emitted, sent to ClusterFinished, when the analysis is complete
        void PreprocFinished(WavesFilePtr waves_file, short elc);
        void CreateLaunchWG(wv::CLUSTER_ID cid);
        void UpdateWavesShown();
        void ElectrodeSet(short elc, WavesFilePtr waves_file);
        void ResetElectrodeGui(WavesFilePtr waves_file);
        void WaveGuiFocus(wv::CLUSTER_ID cid);
        void WaveGuiClosed(wv::CLUSTER_ID cid);

    private:
        void OnActionREXEFileTriggered(bool is_prog=true);
        void OnActionCluFileTriggered(bool is_prog=true);
        void OnActionTCluFileTriggered(bool is_prog=true);
        void WriteCluFile(bool is_prog, bool with_time);
        void OnActionSaveAllTriggered(bool is_prog=true);
        void OnActionSimpleBinaryTriggered(bool is_prog=true);
        void OnActionSimpleASCIITriggered(bool is_prog=true);
        void OnActionRunBatchTriggered(bool is_prog=true);
        void OnActionLoadClassificationTriggered(bool is_prog=true);
        void OnActionLoadWaveformFileTriggered(bool is_prog=true);

        void OnActionSessionTriggered(bool is_prog=true);
        void OnActionKMeansTriggered(bool is_prog=true);
        void OnActionAutoClassTriggered(bool is_prog=true);
        void OnActionSPCTriggered(bool is_prog=true);
        void OnActionKlustaKwikTriggered(bool is_prog=true);
        void OnDisplayParameters(ParamsBase* pars, bool is_prog=true);

#ifndef USE_GSL
        void CheckLicense();
#endif
        wv::CLUSTER_ID ClusterFromFile(InputFilePtr input_file, short elc,
                             ClassificationCPtr cls, wv::CLUSTER_ID cid = wv::DEFAULT_CID);
            //For either loading a classification or an online clustering
        WavesFilePtr LoadDataFile(const std::string& data_file);
        void SaveParameters(const std::string& file_name);

        Ui::MainWindow* ui;
        QLabel* _statusBarLabel;
        WaveSpaceGui* _waveSpaceGui;

        AboutGui* _aboutGui, * _waveSorterProGui;
        ElectrodeGui* _electrodeGui;
        QMdiSubWindow* _electrodeSubWin;
        LogGui* _logGui;
        ParamGui* _paramGui;
        ProfileGui* _profileGui;
        ProgressGui* _progressGui;
#ifdef HAVE_EXTRA
        WaveGridGui* _waveGridGui;
#endif
        Batch* _batch;
        StringList _batchList;
        ChannelBatch* _channelBatch;
        ClusterMediator& _cmed;
        int _electrodeCt; //For keeping track of how many electrodes have been loaded

        std::mutex _mutex;
};

Q_DECLARE_METATYPE(std::string)
Q_DECLARE_METATYPE(WavesFilePtr)

#endif // MAINWINDOW_H
