#include "wavespacegui.h"
#include "ui_wavespacegui.h"
#include "../analysis/threads/mgr/histscattermgr.h"


const int WaveSpaceGui::DEFAULT_NUM_XBINS = 200;
const int WaveSpaceGui::DEFAULT_NUM_YBINS = 200;

WaveSpaceGui::WaveSpaceGui(QWidget *parent)
    : QWidget(parent), ui(new Ui::WaveSpaceGui), _cmed(ClusterMediator::Instance()),
      _waveData(nullptr), _wsgSpec(NULL),
      _histAccept(NULL), _histUnaccept(NULL), _matAccept(NULL), _matUnaccept(NULL), _needNewTemplates(true),
      _numXBins(WaveSpaceGui::DEFAULT_NUM_XBINS),  _numYBins(WaveSpaceGui::DEFAULT_NUM_YBINS),
      _prevNumXBins(WaveSpaceGui::DEFAULT_NUM_XBINS),  _prevNumYBins(WaveSpaceGui::DEFAULT_NUM_YBINS),
      _scaleSpec(new ScaleSpec()), _waitForEllipse(false)
{
    ui->setupUi(this);
    ui->wgtWaves->SetBackground( QColor(31,31,31) );
    SetDrawButtons(false);
    ui->sldTime->setEnabled(false);
    ui->spnSD->setValue(BiEllipse::DEFAULT_ELLIPSE_SD);
#if defined(Q_WS_MAC) || defined Q_OS_MACX || defined(Q_OS_WIN)
    ui->rbtnAutoClass->setEnabled(false);
#endif

    connect(ui->wgtHistogram, SIGNAL(DiskMoved(int,wv::float_t,wv::float_t)), this, SLOT(HistDiskMoved(int,wv::float_t,wv::float_t)));
    connect(ui->wgtHistogram, SIGNAL(ResetZoomMode()), this, SLOT(ResetZoomMode()));
    connect(ui->wgtHistogram, SIGNAL(RecalculateScatter()), this, SLOT(RecalculateScatter()));
    connect(ui->wgtHistogram, SIGNAL(WheelStep(bool)), this, SLOT(HistWheelMoved(bool)));
    connect(ui->wgtHistogram, SIGNAL(SendScrubIndex(int)), ui->wgtWaves, SLOT(ReceiveScrubIndex(int)));
    connect(ui->wgtHistogram, SIGNAL(ToggleWaveform(int)), this, SLOT(ReceiveToggle(int)));
    connect(ui->wgtHistogram, SIGNAL(SendDragEvent()), this, SLOT(ReceiveDragEvent()));

    connect(ui->sldTime, SIGNAL(sliderMoved(int)), this, SLOT(TimeSliderMoved(int)));
    connect(ui->sldHistX, SIGNAL(sliderMoved(int)), this, SLOT(HistSliderMoved(int)));

    connect(ui->sldWaves, SIGNAL(sliderMoved(int)), this, SLOT(WavesSliderMoved(int)));
    connect(ui->wgtWaves, SIGNAL(RangeChange(int,int)), this, SLOT(WavesRangeChanged(int,int)));
    connect(ui->wgtWaves, SIGNAL(SendScrubIndex(int)), ui->wgtHistogram, SLOT(ReceiveScrubIndex(int)));
    connect(ui->wgtWaves, SIGNAL(ToggleWaveform(int)), this, SLOT(ReceiveToggle(int)));

    connect(ui->sldWindow, SIGNAL(sliderMoved(int)), this, SLOT(WindowSliderMoved(int)));
    connect(ui->sldNumShown, SIGNAL(sliderMoved(int)), this, SLOT(NumShownSliderMoved(int)));
    connect(ui->edWindow, SIGNAL(editingFinished()), this, SLOT(EdWindowChanged()));
    connect(ui->edNumShown, SIGNAL(editingFinished()), this, SLOT(EdNumShownChanged()));

    SetMovieGui(false, false);
    connect(&_timer, SIGNAL(timeout()), this, SLOT(NextFrame()));
}

WaveSpaceGui::~WaveSpaceGui()
{
    delete ui;
}

//###Should make this a (static) method of CluSpecPtr or something
void WaveSpaceGui::InitializeSpecs(wv::CLUSTER_ID cid)
{
    CluSpecPtr clu_spec = _cmed.GetCluSpecPtr(cid);
    WSGSpec& wsg_spec = clu_spec->GetWSGSpec();
    WaveData* wave_data = _cmed.GetWaveData(cid);

    // *** Set up _cluSpec
    clu_spec->SetWaveData(wave_data);
    clu_spec->SetSelectedSamplesBegin(0);
    if (clu_spec->Transform()==wv::USER)
    {
        wv::TRANSFORM_ID old_tid = clu_spec->TransformId(),
                new_tid = wave_data->CreateUserTransform(old_tid);
        clu_spec->SetTransformId(new_tid);
    }
    else if (clu_spec->TransformId()==wv::NO_TRANSFORM_ID)
        clu_spec->SetTransform(wv::RAW, wv::DEFAULT_TID);
    int num_samples = wave_data->NumSamples( clu_spec->Transform(), clu_spec->TransformId() );
    clu_spec->SetSelectedSamplesEnd(num_samples);
    clu_spec->InitToggledWaves();
    // *** End setup _cluSpec

    // ***Now set up and update wsg_spec***
    wv::TRANSFORM transform = clu_spec->Transform();
    wsg_spec.SetGuiTransform(transform);
    wv::TRANSFORM_ID tid = clu_spec->TransformId();

        //Set the transforms
    if (wsg_spec.Transform(WSGSpec::X)==wv::UNKNOWN_TRANSFORM
            || wsg_spec.Transform(WSGSpec::Y)==wv::UNKNOWN_TRANSFORM)
    {
        wsg_spec.SetTransform(WSGSpec::X, wv::TIMESTAMP, 0);
        wsg_spec.SetTransform(WSGSpec::Y, wv::RAW, 0);
    }

    //Set the properties governing display of the transforms
    WSGSpec::AXIS active_axis;
    if (cid==wv::DEFAULT_CID)
    {
        active_axis = WSGSpec::Y;
        wsg_spec.SetSample(WSGSpec::X, 0);
        wsg_spec.SetSample(WSGSpec::Y, 0);
        wsg_spec.SetActiveAxis(WSGSpec::Y);
        wsg_spec.SetWavesTransform( wsg_spec.Transform(WSGSpec::Y) );
    }
    else
    {
        active_axis = transform==wsg_spec.Transform(WSGSpec::X) && tid==wsg_spec.TransformId(WSGSpec::X)
                ? WSGSpec::X : WSGSpec::Y;
        wsg_spec.SetSample(WSGSpec::X, clu_spec->GetWSGSpec().Sample(WSGSpec::X));
        wsg_spec.SetSample(WSGSpec::Y, clu_spec->GetWSGSpec().Sample(WSGSpec::Y));
        wsg_spec.SetActiveAxis(active_axis);
        wsg_spec.SetWavesTransform( wsg_spec.Transform(active_axis) );
    }
    wsg_spec.SetSamplesBegin(0);
    wsg_spec.SetSamplesEnd( wave_data->NumSamples(
                                wsg_spec.Transform(active_axis), wsg_spec.TransformId(active_axis)) );

    //Final details for _wsgSpec
    wsg_spec.InitWaveShown();
    wsg_spec.SetWindowLimits( ui->edWindow->text().toInt() );
    // ***End setting up _wsgSpec***
}

void WaveSpaceGui::Setup(wv::CLUSTER_ID cid)
{
    //This function is called every time a different WaveGui gets focus or a file gets
    //loaded

    if (cid==wv::DEFAULT_CID)
        _cmed.InitSpecs(wv::DEFAULT_CID);
    else
        _cmed.CopyCID(cid);

    InitializeSpecs(wv::DEFAULT_CID);

    _cluSpec = _cmed.GetCluSpecPtr(wv::DEFAULT_CID);
    _wsgSpec = &_cluSpec->GetWSGSpec();
    _waveData = _cmed.GetWaveData(wv::DEFAULT_CID);
    int num_records = _waveData->NumRecords(),
        num_samples = _waveData->NumSamples( _wsgSpec->WavesTransform(), _wsgSpec->WavesTransformId() );
    if (_waveData->NumUserTransforms()==0)
        _waveData->SetUTsFromSignatures();

    const wv_vector_int* clus_accepted;
    if (cid==wv::DEFAULT_CID)
    {
        clus_accepted = nullptr;
        _cmed.SetClusAccepted(cid, nullptr);
    }
    else
        clus_accepted = _cmed.ClusAccepted(cid); // *Not* wv::DEFAULT_CID, take this from the incoming
            //clustering

    //Title and misc. gui features
    std::string clus_str = (cid==wv::DEFAULT_CID)
            ? "" : "clustering " + std::to_string(cid) + ", ";
    setWindowTitle( ("WaveSpace, " + clus_str + _waveData->FileName()
                     + ", electrode " + std::to_string(_cmed.GetElectrode(wv::DEFAULT_CID))).c_str() );
    _needNewTemplates = true;
    MakeBAandColors(num_records);
    ui->sldWindow->setMaximum(num_records-1);

    //ScanWaveWidget range effectively taken care of in InitSpecs

    //Time
    ui->sldTime->setMinimum( _wsgSpec->Timestamp(0) );
    ui->sldTime->setMaximum( _wsgSpec->Timestamp(num_records-1) );
    ui->sldTime->setValue(0);
    ui->sldTime->setEnabled( _wsgSpec->HistLimitMode()==WSGSpec::TIMESTAMP );
    ui->cboxTime->setChecked( _wsgSpec->HistLimitMode()==WSGSpec::TIMESTAMP );

    //Initializing the widgets
    ui->wgtWaves->InitGraphics(_cluSpec);
    ui->wgtHistogram->InitGraphics(_cluSpec, _scaleSpec); //Could just pass the cid, but this seems
        //clearer
    ui->wgtHistogram->ResetScaleSpecVals();
    ui->sldWaves->setMaximum( num_samples-1 );

    //Transform radio buttons and user lists
    SetUserControls( _waveData->NumUserTransforms()>0 );
    ui->rbtnHeight->setEnabled( _waveData->HaveTransform(wv::HEIGHT) );
    ui->rbtnWidth->setEnabled( _waveData->HaveTransform(wv::WIDTH) );
    ui->rbtnSlope->setEnabled( _waveData->HaveTransform(wv::SLOPE) );
    ui->rbtnSlopeD->setEnabled( _waveData->HaveTransform(wv::SLOPE_D) );
#ifdef USE_GSL
    ui->rbtnEnergy->setEnabled(false);
#else
    ui->rbtnEnergy->setEnabled( _waveData->HaveTransform(wv::ENERGY) );
#endif
    ui->rbtnPCA->setEnabled( _waveData->HaveTransform(wv::PCA) );
#ifdef USE_GSL
    ui->rbtnSpline->setEnabled( _waveData->HaveTransform(wv::SPLINE) );
    ui->rbtnSplineD->setEnabled( _waveData->HaveTransform(wv::SPLINE_D) );
    ui->rbtnFourier->setEnabled( _waveData->HaveTransform(wv::FOURIER) );
    ui->rbtnFourierD->setEnabled( _waveData->HaveTransform(wv::FOURIER_D) );
#else
    ui->rbtnSpline->setEnabled(false);
    ui->rbtnSplineD->setEnabled(false);
    ui->rbtnFourier->setEnabled(false);
    ui->rbtnFourierD->setEnabled(false);
    ui->rbtnSpline->hide();
    ui->rbtnSplineD->hide();
    ui->rbtnFourier->hide();
    ui->rbtnFourierD->hide();
#endif
    ui->rbtnWavelets->setEnabled( _waveData->HaveTransform(wv::WAVELET) );
    ui->rbtnRawD->setEnabled( _waveData->HaveTransform(wv::RAW_D) );
    ui->rbtnPCAD->setEnabled( _waveData->HaveTransform(wv::PCA_D) );
    ui->rbtnWaveletsD->setEnabled( _waveData->HaveTransform(wv::WAVELET_D) );

    //Which zoom type
    switch (_wsgSpec->HistZoomMode())
    {
        case WSGSpec::AXIS_ZOOM:
            on_rbtnHistAxis_clicked();
            break;
        case WSGSpec::RESOLUTION:
            on_rbtnHistResolution_clicked();
            break;
        case WSGSpec::LUMINANCE:
            on_rbtnHistLuminance_clicked();
            break;
        case WSGSpec::DISK_SIZE:
            break;
    }

    //Which cursor
    ui->rbtnBar->setChecked(true);
    _wsgSpec->SetHistCursorMode(WSGSpec::BAR);
    ui->wgtWaves->SetUseDashes(false);
    for (int i=0; i<WSGSpec::NUM_DISKS; ++i)
        ui->wgtHistogram->ResetDisks(i);

    //Waves view radio button
    switch (_wsgSpec->ActiveAxis())
    {
        case WSGSpec::X:
            ui->rbtnSetWavesFromXAxis->setChecked(true);
            break;
        case WSGSpec::Y:
            ui->rbtnSetWavesFromYAxis->setChecked(true);
            break;
    }

    //Set selected transform button
    switch (_wsgSpec->Transform(WSGSpec::Y))
    {
        case wv::USER:
            ui->rbtnUser->setChecked(true);
            break;
        case wv::HEIGHT:
            ui->rbtnHeight->setChecked(true);
            break;
        case wv::WIDTH:
            ui->rbtnWidth->setChecked(true);
            break;
        case wv::SLOPE:
            ui->rbtnSlope->setChecked(true);
            break;
        case wv::SLOPE_D:
            ui->rbtnSlopeD->setChecked(true);
            break;
        case wv::ENERGY:
            ui->rbtnEnergy->setChecked(true);
            break;
        case wv::TIMESTAMP:
            ui->rbtnTimestamps->setChecked(true);
            break;
        case wv::RAW:
            ui->rbtnRaw->setChecked(true);
            break;
        case wv::RAW_D:
            ui->rbtnRawD->setChecked(true);
            break;
        case wv::PCA:
            ui->rbtnPCA->setChecked(true);
            break;
        case wv::PCA_D:
            ui->rbtnPCAD->setChecked(true);
            break;
#ifdef USE_GSL
        case wv::SPLINE:
            ui->rbtnSpline->setChecked(true);
            break;
        case wv::SPLINE_D:
            ui->rbtnSpline->setChecked(true);
            break;
        case wv::FOURIER:
            ui->rbtnFourier->setChecked(true);
            break;
        case wv::FOURIER_D:
            ui->rbtnFourierD->setChecked(true);
            break;
#endif
        case wv::WAVELET:
            ui->rbtnWavelets->setChecked(true);
            break;
        case wv::WAVELET_D:
            ui->rbtnWaveletsD->setChecked(true);
            break;
        default:
            break;
    }

    //Set selected cluster button
    switch (_cluSpec->ClusterMethod())
    {
        case wv::MANUAL:
            ui->rbtnManual->setChecked(true);
            break;
        case wv::AUTOCLASS:
            ui->rbtnAutoClass->setChecked(true);
            break;
        case wv::KMEANS:
            ui->rbtnKMeans->setChecked(true);
            break;
        case wv::SPC:
            ui->rbtnSPC->setChecked(true);
            break;
        case wv::KK:
            ui->rbtnKK->setChecked(true);
            break;
        default:
            _cluSpec->SetClusterMethod(wv::KMEANS);
            _cluSpec->SetClusterParameters( ClusterParamsBase::Create(wv::KMEANS) );
            ui->rbtnKMeans->setChecked(true);
            break;
    }

    //Draw line controls and the drawlines themselves
    SetDrawButtons( _cluSpec->ClusterMethod()==wv::MANUAL );
    ui->cmbDrawLines->clear();
    int num_polylines = _cluSpec->Polylines().size();
    for (int i=0; i<num_polylines; ++i)
    {
        std::string name = _cluSpec->Polylines().at(i)->Name();
        ui->cmbDrawLines->addItem(name.c_str());
    }
    int num_circles = _cluSpec->Circles().size();
    for (int i=0; i<num_circles; ++i)
    {
        std::string name = _cluSpec->Circles().at(i)->Name();
        ui->cmbDrawLines->addItem(name.c_str());
    }
    ui->btnScrub->setEnabled(clus_accepted);

    //Count the number of accepted waveforms
    int num_accepted = 0;
    if (clus_accepted)
        for (int i=0; i<num_records; ++i)
            if (wv_vector_int_get(clus_accepted,i) == wv::ACCEPT_CLUSTER) ++num_accepted;

    connect(_waveData, SIGNAL(EnableTransform(wv::TRANSFORM,short)), this, SLOT(EnableTransform(wv::TRANSFORM,short)));

    UpdateAll();

    //Purge unused transforms--NB transforms still being calculated at this point won't be
    //deleted until the next call.
    SafeUnload();
}

LabelWidget* WaveSpaceGui::AxisLabel(WSGSpec::AXIS axis) const
{
    LabelWidget* label(NULL);
    switch (axis)
    {
        case WSGSpec::X:
            label = ui->wgtHistLabelX;
            break;
        case WSGSpec::Y:
            label = ui->wgtHistLabelY;
            break;
    }
    return label;
}

//NB this must be called only *after* cid has been copied into the WaveSpaceGui
void WaveSpaceGui::ReceiveClusters(wv::CLUSTER_ID cid)
{
    const wv_vector_int* clus_accepted = _cmed.ClusAccepted(cid);
    int num_records = clus_accepted->size;
    wv_vector_int* new_clus_accepted = wv_vector_int_alloc(num_records);
    wv_vector_int_memcpy(new_clus_accepted, clus_accepted);
    _cmed.SetClusAccepted(wv::DEFAULT_CID, new_clus_accepted);
    MakeBAandColors(num_records);
    MakeTemplates();
    ui->btnScrub->setEnabled(true);    
    _needNewTemplates = true;
    UpdateAll();
}
void WaveSpaceGui::HistWheelMoved(bool is_forward)
{
    switch ( _wsgSpec->HistZoomMode() )
    {
        case WSGSpec::AXIS_ZOOM:
        {
            wv::float_t axis_zoom = _wsgSpec->HistAxisZoom();
            axis_zoom *= is_forward ? WSGSpec::ZOOM_INCREMENT : 1.0/WSGSpec::ZOOM_INCREMENT;
            axis_zoom = qMax((wv::float_t)1.0, axis_zoom);
            if (axis_zoom > 1.01 || is_forward)
            {
                _numXBins = (is_forward ? _numXBins*WSGSpec::ZOOM_INCREMENT
                                        : _numXBins/WSGSpec::ZOOM_INCREMENT) + 0.5;
                _numYBins = (is_forward ? _numYBins*WSGSpec::ZOOM_INCREMENT
                                        : _numYBins/WSGSpec::ZOOM_INCREMENT) + 0.5;
                //Zooming in is not very useful without zooming the resolution too.
            }
            _wsgSpec->SetHistAxisZoom(axis_zoom);
            UpdateHistogram();
            UpdateHistWidget(); //To update the slider range
            break;
        }
        case WSGSpec::RESOLUTION:
            _prevNumXBins = _numXBins = (is_forward ? _numXBins*WSGSpec::ZOOM_INCREMENT
                                                    : _numXBins/WSGSpec::ZOOM_INCREMENT) + 0.5;
            _prevNumYBins = _numYBins = (is_forward ? _numYBins*WSGSpec::ZOOM_INCREMENT
                                                    : _numYBins/WSGSpec::ZOOM_INCREMENT) + 0.5;
            UpdateHistogram();
            UpdateHistWidget(); //For ScaleSpec
            break;
        case WSGSpec::LUMINANCE:
        {
            wv::float_t lum_scale = _wsgSpec->HistLumScale();
            lum_scale = is_forward ? lum_scale/WSGSpec::ZOOM_INCREMENT
                                   : lum_scale*WSGSpec::ZOOM_INCREMENT;
            _wsgSpec->SetHistLumScale(lum_scale);
            _wsgSpec->SetHistPrevLumScale(lum_scale);
            break;
        }
        case WSGSpec::DISK_SIZE:
            int i = ui->wgtHistogram->SelectedDisk();
            int hist_disk_size = _wsgSpec->HistDiskSize(i);
            int new_size = (is_forward ?hist_disk_size*WSGSpec::ZOOM_INCREMENT
                                      : hist_disk_size/WSGSpec::ZOOM_INCREMENT) + 0.5;
            if (new_size==hist_disk_size) new_size = is_forward ? hist_disk_size+2 : 0; //When rounding doesn't increase pixel size
            _wsgSpec->SetHistDiskSize(i, new_size);
            ResetDiskCursors();
            UpdateWaves();
            UpdateWavesWidget();
            break;
    }
    ui->wgtHistogram->update();
}
void WaveSpaceGui::EnableTransform(wv::TRANSFORM transform, short electrode)
{
    if (electrode != _cmed.GetWaveData(wv::DEFAULT_CID)->Electrode()) return;
    if (!_waveData->HaveTransform(transform)) return;
        //This is possible if a new clustering has been loaded before transforms from the previous
        //one have finished calculating
    switch (transform)
    {
        case wv::USER:
            UpdateUTContents();
            SetUserControls(true);
            break;
        case wv::HEIGHT:
            ui->rbtnHeight->setEnabled(true);
            break;
        case wv::WIDTH:
            ui->rbtnWidth->setEnabled(true);
            break;
        case wv::SLOPE:
            ui->rbtnSlope->setEnabled(true);
            break;
        case wv::SLOPE_D:
            ui->rbtnSlopeD->setEnabled(true);
            break;
        case wv::ENERGY:
            ui->rbtnEnergy->setEnabled(true);
            break;
        case wv::PCA:
            ui->rbtnPCA->setEnabled(true);
            break;
        case wv::PCA_D:
            ui->rbtnPCAD->setEnabled(true);
            break;
#ifdef USE_GSL
        case wv::SPLINE:
            ui->rbtnSpline->setEnabled(true);
            break;
        case wv::SPLINE_D:
            ui->rbtnSplineD->setEnabled(true);
            break;
        case wv::FOURIER:
            ui->rbtnFourier->setEnabled(true);
            break;
        case wv::FOURIER_D:
            ui->rbtnFourierD->setEnabled(true);
            break;
#endif
        case wv::WAVELET:
            ui->rbtnWavelets->setEnabled(true);
            break;
        case wv::WAVELET_D:
            ui->rbtnWaveletsD->setEnabled(true);
            break;
        default:
            break;
    }
}
void WaveSpaceGui::ReceiveDragEvent()
{
    UpdateHistWidget();
    if (_wsgSpec->HistCursorMode()==WSGSpec::BAR) DrawHistWindow();
}
void WaveSpaceGui::ReceiveToggle(int scrub_index)
{
    ToggleWave(scrub_index);
    MakeBAandColors( _cmed.GetWaveData(wv::DEFAULT_CID)->NumRecords() );
    UpdateAll();
}
void WaveSpaceGui::StopMovie()
{
    if ( _wsgSpec->MovieState() == WSGSpec::PLAYING) OnBtnPlayClicked();
}

void WaveSpaceGui::resizeEvent(QResizeEvent*)
{
    DrawScanline();
    if ( _wsgSpec->HistCursorMode()==WSGSpec::BAR) DrawHistWindow();
}

void WaveSpaceGui::OnCboxTimeClicked(bool value, bool)
{
    ui->sldTime->setEnabled(value);
    ui->sldHistX->setEnabled(!value);
    _wsgSpec->SetHistLimitMode( value ? WSGSpec::TIMESTAMP : WSGSpec::TRANSFORM );
    int window_size = ui->edWindow->text().toInt();
    _wsgSpec->SetWindowLimits(window_size);
    UpdateAll();
}
void WaveSpaceGui::TimeSliderMoved(int value)
{
    _wsgSpec->SetCurrentTime(value);
    UpdateTimeSlider();
    ui->edSliderValue->setText( std::to_string(value).c_str() );
    UpdateAll();
}

void WaveSpaceGui::HistSliderMoved(int value)
{
    try
    {
        if (ui->rbtnSetWavesFromRaw->isChecked()) //Otherwise X must already be the active slider
            _wsgSpec->SetActiveAxis(WSGSpec::Y);
        _wsgSpec->SetCurrentValue(value);
        int window_size = ui->edWindow->text().toInt();
        _wsgSpec->SetWindowLimits(window_size);
            //Setting the range of viewable records, determined by the width of the window attached to the active
            //slider in _wgtHistogram

        ui->edSliderValue->setText( std::to_string(value).c_str() );

        DrawHistWindow();
        ui->wgtHistogram->update();
            //We don't call UpdateHistWidget because redrawing the entire histogram can be unnecessary and
            //is expensive.
        UpdateWaves();
        UpdateWavesWidget();
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}

void WaveSpaceGui::HistDiskMoved(int selected_disk, wv::float_t x, wv::float_t y) //slot
{
    try
    {
        wv::float_t hist_disk_size = (wv::float_t)_wsgSpec->HistDiskSize(selected_disk);
        wv::float_t xdist = _wsgSpec->Range(WSGSpec::X) * hist_disk_size/(wv::float_t)ui->wgtHistogram->width(),
                ydist = _wsgSpec->Range(WSGSpec::Y) * hist_disk_size/(wv::float_t)ui->wgtHistogram->HistHeight();
        _wsgSpec->SetDiskCursor(selected_disk, x, y,
                                xdist / _wsgSpec->HistAxisZoom(),
                                ydist / _wsgSpec->HistAxisZoom());
            //NB this could be done directly from HistWidget...

        ui->edSliderValue->setText("");

        ui->wgtHistogram->update();
            //We don't call UpdateHistWidget because redrawing the entire histogram can be unnecessary and
            //is expensive.
        UpdateWaves();
        UpdateWavesWidget();
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}
void WaveSpaceGui::ResetZoomMode()
{
    if (ui->rbtnHistAxis->isChecked()) _wsgSpec->SetHistZoomMode(WSGSpec::AXIS_ZOOM);
    else if (ui->rbtnHistResolution->isChecked()) _wsgSpec->SetHistZoomMode(WSGSpec::RESOLUTION);
    else if (ui->rbtnHistLuminance->isChecked()) _wsgSpec->SetHistZoomMode(WSGSpec::LUMINANCE);
}
void WaveSpaceGui::RecalculateScatter()
{
    _numXBins = ui->wgtHistogram->width();
    _numYBins = ui->wgtHistogram->HistHeight();
    UpdateHistogram();
}
void WaveSpaceGui::WavesSliderMoved(int value)
{
    try
    {
        WSGSpec::AXIS active_axis = _wsgSpec->ActiveAxis();
        MatCPtr all_waves = _waveData->Waveforms(
                    _wsgSpec->Transform(active_axis), _wsgSpec->TransformId(active_axis) );
        if ((uint)value >= all_waves->size2)
            throw BadGuiOp("WaveSpaceGui::WavesSliderMoved", "Slider value out of range");
            //This can happen with the user vector
        _wsgSpec->SetSample(active_axis, value);
        FinishDrawline();
        ui->wgtHistogram->ResetScaleSpecVals();

        //Set histogram labels
        wv::TRANSFORM transform = _wsgSpec->Transform(active_axis);
        wv::TRANSFORM_ID tid = _wsgSpec->TransformId(active_axis);
        std::string label_text = _waveData->TransformLabel(transform, value, tid);
        AxisLabel(active_axis)->SetText(label_text, active_axis);

        //Call relevant update functions
        UpdateHistogram();
        if ( _wsgSpec->HistCursorMode()==WSGSpec::DISK)
        {
            ResetDiskCursors();
            UpdateWaves();
        }
        else if ( _wsgSpec->ActiveAxis()==WSGSpec::X )
            ui->sldHistX->setValue( std::numeric_limits<int>::min() ); //Neanderthal approach
        UpdateHistWidget( _wsgSpec->ActiveAxis()==WSGSpec::X );
        UpdateWavesWidget(); //For the scanline/disks

        if (_wsgSpec->GuiTransform() == wv::USER)
            OnLsUTContentsNewIndex(value);
    }
    catch (std::exception& e)
    {
        Log(e.what());
    }
}
void WaveSpaceGui::WavesRangeChanged(int begin, int end) //Gets called after a gui operation on the Waves
    //widget
{
    int current_value = _wsgSpec->Sample( _wsgSpec->ActiveAxis() ),
        current_min = ui->sldWaves->minimum(),
        current_max = ui->sldWaves->maximum();
    wv::float_t value_pct = (wv::float_t)(current_value - current_min) / (wv::float_t)(current_max - current_min);
    int new_value = value_pct * (end - begin - 1) + begin + 0.5;

    ui->sldWaves->setRange(begin, end-1);
    new_value = qBound(begin, new_value, end-1);
    if ( !ui->rbtnSetWavesFromRaw->isChecked() )
    {
        WavesSliderMoved(new_value);
        ui->sldWaves->setValue(new_value);
    }
    else
        ui->sldWaves->setValue(0);
}
void WaveSpaceGui::EdWindowChanged()
{
    int old_num = ui->sldWindow->value();
    bool is_valid;
    int new_num = ui->edWindow->text().toInt(&is_valid);
    if (!is_valid)
    {
        ui->edWindow->setText( std::to_string(old_num).c_str() );
        return;
    }
    new_num = qBound(1, new_num, _waveData->NumRecords());
    ui->sldWindow->setValue(new_num);
    if ( _wsgSpec->HistCursorMode()==WSGSpec::BAR ) DrawHistWindow();
    UpdateWaves();
    UpdateWavesWidget();
}
void WaveSpaceGui::EdNumShownChanged()
{
    int old_num = ui->sldNumShown->value();
    bool is_valid;
    int new_num = ui->edNumShown->text().toInt(&is_valid);
    if (!is_valid)
    {
        ui->edNumShown->setText( std::to_string(old_num).c_str() );
        return;
    }
    new_num = qBound(1, new_num, ui->sldWindow->value());
    ui->sldNumShown->setValue(new_num);
    UpdateWaves();
    UpdateWavesWidget();
}
void WaveSpaceGui::WindowSliderMoved(int value)
{
    ui->edWindow->setText( std::to_string(value).c_str() ); //Does not emit editingFinished()
    _wsgSpec->SetWindowLimits(value);
    if (_wsgSpec->HistLimitMode()==WSGSpec::TIMESTAMP)
    {
        UpdateTimeSlider();
        UpdateHistogram();
    }
    UpdateHistWidget();
    UpdateWaves();
    UpdateWavesWidget();
}
void WaveSpaceGui::NumShownSliderMoved(int value)
{
    ui->edNumShown->setText( std::to_string(value).c_str() ); //Does not emit editingFinished()
    UpdateWaves();
    UpdateWavesWidget();
}

void WaveSpaceGui::on_edNumShown_editingFinished()
{
    UpdateWaves();
    UpdateWavesWidget();
}

void WaveSpaceGui::on_rbtnHistAxis_clicked()
{
    _wsgSpec->SetHistZoomMode(WSGSpec::AXIS_ZOOM);
}
void WaveSpaceGui::on_rbtnHistResolution_clicked()
{
    _wsgSpec->SetHistZoomMode(WSGSpec::RESOLUTION);
}
void WaveSpaceGui::on_rbtnHistLuminance_clicked()
{
    _wsgSpec->SetHistZoomMode(WSGSpec::LUMINANCE);
}
void WaveSpaceGui::on_btnScatter_clicked()
{
    std::string text = ui->btnScatter->text().toStdString();
    if (text=="Scatter")
    {
       _wsgSpec->SetIsScatter(true);
        _numXBins = ui->wgtHistogram->width();
        _numYBins = ui->wgtHistogram->HistHeight();
        _wsgSpec->SetHistLumScale(0.0); //I.e., all pixel values greater than 0 are white
        ui->btnScatter->setText("Histogram");
    }
    else //"Histogram"
    {
        _wsgSpec->SetIsScatter(false);
        _numXBins = _prevNumXBins;
        _numYBins = _prevNumYBins;
        _wsgSpec->SetHistLumScale( _wsgSpec->HistPrevLumScale() );
        ui->btnScatter->setText("Scatter");
    }

    UpdateHistogram();
    ui->wgtHistogram->update();
}
void WaveSpaceGui::on_btnMovieGui_clicked()
{
    if (ui->layMovie->isEnabled())
    {
        if ( _wsgSpec->MovieState()==WSGSpec::PLAYING ) OnBtnPlayClicked();
        SetMovieGui(false);
        ui->btnMovieGui->setText("Movie");
    }
    else
    {
        SetMovieGui(true);
        ui->btnMovieGui->setText("Close Movie");
    }
}

void WaveSpaceGui::OnBtnPlayClicked(bool)
{
    switch ( _wsgSpec->MovieState() )
    {
        case WSGSpec::STOPPED:
        {
            ui->btnPlay->setText("Stop");
            int interval = 1000.0 / (wv::float_t)ui->spnFrames->value() + 0.5;
            _timer.start(interval);
            _wsgSpec->SetMovieState(WSGSpec::PLAYING);
            break;
        }
        case WSGSpec::PLAYING:
            ui->btnPlay->setText("Play");
            _timer.stop();
            _wsgSpec->SetMovieState(WSGSpec::STOPPED);
            break;
    }
}
void WaveSpaceGui::NextFrame()
{
    wv::float_t current, window_end, axis_max;
    switch (_wsgSpec->HistLimitMode())
    {
        case WSGSpec::TRANSFORM:
            current = _wsgSpec->CurrentValue();
            window_end = _wsgSpec->WindowEnd();
            axis_max = _wsgSpec->Max(WSGSpec::X);
            break;
        case WSGSpec::TIMESTAMP:
            if ((int)_wsgSpec->CurrentTime() == -1)
                _wsgSpec->SetCurrentTime( _wsgSpec->Timestamp(0));
            current = _wsgSpec->CurrentTime();
            window_end = _wsgSpec->TimeEnd();
            axis_max = _wsgSpec->Timestamp(-1);
            break;
    }

    wv::float_t base_step = window_end - current,
            next = current + base_step/(wv::float_t)ui->spnWaveDuration->value();

    if (next < axis_max - base_step)
        switch (_wsgSpec->HistLimitMode())
        {
            case WSGSpec::TRANSFORM:
                switch (_wsgSpec->HistLimitMode())
                {
                    case WSGSpec::TRANSFORM:
                        ui->sldHistX->setValue(next);
                        break;
                    case WSGSpec::TIMESTAMP:
                        ui->sldTime->setValue(next);
                        break;
                }
                HistSliderMoved(next);
                break;
            case WSGSpec::TIMESTAMP:
                ui->sldTime->setValue(next);
                TimeSliderMoved(next);
                break;
        }
    else
        on_btnPlay_clicked();
}

void WaveSpaceGui::OnRbtnSetWavesFromXAxisClicked(bool)
{
    _wsgSpec->SetWavesTransform( _wsgSpec->Transform(WSGSpec::X) );
    _wsgSpec->SetActiveAxis(WSGSpec::X);
    UpdateWaves();
    UpdateWavesWidget(true);
}
void WaveSpaceGui::OnRbtnSetWavesFromYAxisClicked(bool)
{
    _wsgSpec->SetWavesTransform( _wsgSpec->Transform(WSGSpec::Y) );
    _wsgSpec->SetActiveAxis(WSGSpec::Y);
    UpdateWaves();
    UpdateWavesWidget(true);
}
void WaveSpaceGui::OnRbtnSetWavesFromRawClicked(bool)
{
    _wsgSpec->SetWavesTransform(wv::RAW);
    ui->sldWaves->setEnabled(false);
    if ( _wsgSpec->HistCursorMode()==WSGSpec::DISK ) ResetDiskCursors();
    UpdateWaves();
    UpdateWavesWidget(true);
}

void WaveSpaceGui::OnBtnNewUserTransformClicked(bool)
{
    int num_user_transforms = _waveData->NumUserTransforms();
    ui->spnUserTransform->setMaximum(num_user_transforms++);
    _waveData->CreateUserTransform();
    ui->spnUserTransform->setValue( num_user_transforms-1 );
}
void WaveSpaceGui::OnBtnSaveUserTransformsClicked(bool)
{
    ProfileOut profile_out("", ParamVec());
    profile_out.Save();
    profile_out.Close();
}
void WaveSpaceGui::OnBtnAddToUserClicked(bool)
{
    WSGSpec::AXIS active_axis = _wsgSpec->ActiveAxis();
    wv::TRANSFORM_ID user_tid = wv::NO_TRANSFORM_ID;
    if (_waveData->NumUserTransforms()==0)
    {
        user_tid = _waveData->CreateUserTransform();
        _wsgSpec->SetGuiTransformId(0);
        _wsgSpec->SetUserTCIndex(0);
        _wsgSpec->SetUserTidIndex(0);
        ui->spnUserTransform->setMaximum(0);
    }
    else
        user_tid = _wsgSpec->UserTid();
    int sample = ui->sldWaves->value();
    wv::TRANSFORM transform = _wsgSpec->Transform(active_axis);
    _waveData->AddToUser( user_tid, _wsgSpec->HistView(active_axis), transform, sample, true );

    UpdateUTContents();
    SetUserControls(true);
}
void WaveSpaceGui::OnBtnRemoveFromUserClicked(bool)
{
    if (_wsgSpec->ActiveTransform() != wv::USER) return;
    wv::TRANSFORM_ID tid = _wsgSpec->GuiTransformId();
    int sample = ui->sldWaves->value();
    if (tid==wv::DEFAULT_TID) throw BadTID("WaveSpaceGui::OnBtnRemoveFromUserClicked", tid);
    _waveData->RemoveFromUser(tid, sample);
    if (_waveData->UserSize(tid) == 0) //When the last vector is removed, the transform is removed
    {
        _wsgSpec->SetTransform(_wsgSpec->ActiveAxis(), wv::RAW);
        ui->spnUserTransform->setValue(0);
        _wsgSpec->SetUserTidIndex(0);
        _wsgSpec->SetWavesTransform(wv::RAW);
        ui->rbtnRaw->setChecked(true);

        int num_user_transforms = _waveData->NumUserTransforms();
        if (num_user_transforms==1) SetUserControls(false);
        _waveData->RemoveUserTransform(tid);
        --num_user_transforms;
        ui->spnUserTransform->setMaximum(num_user_transforms-1);
    }
    else
        ui->sldWaves->setValue( qMax(0, sample-1) );
    UpdateAll(true);
}

void  WaveSpaceGui::on_rbtnBar_clicked()
{
    ui->sldHistX->setEnabled(true);
    ui->wgtWaves->SetUseDashes(false);
    SetHistCursor(WSGSpec::BAR);
    MakeBAandColors(_waveData->NumRecords()); //Resets colors to white/green
}
void  WaveSpaceGui::on_rbtnDisk_clicked()
{
    ui->sldHistX->setEnabled(false);
    ui->wgtWaves->SetUseDashes(true);
    SetHistCursor(WSGSpec::DISK);
}

void WaveSpaceGui::OnBtnSetHistAxisXClicked(bool)
{
    SetHistAxis(WSGSpec::X);
}
void WaveSpaceGui::OnBtnSetHistAxisYClicked(bool)
{
    SetHistAxis(WSGSpec::Y);
}
void WaveSpaceGui::OnRbtnUserClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::USER);
}
void WaveSpaceGui::OnRbtnTimestampsClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::TIMESTAMP);
}
void WaveSpaceGui::OnRbtnHeightClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::HEIGHT);
}
void WaveSpaceGui::OnRbtnWidthClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::WIDTH);
}
void WaveSpaceGui::OnRbtnSlopeClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::SLOPE);
}
void WaveSpaceGui::OnRbtnSlopeDClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::SLOPE_D);
}
void WaveSpaceGui::OnRbtnEnergyClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::ENERGY);
}
void WaveSpaceGui::OnRbtnRawClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::RAW);
}
void WaveSpaceGui::OnRbtnRawDClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::RAW_D);
}
void WaveSpaceGui::OnRbtnPCAClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::PCA);
}
void WaveSpaceGui::OnRbtnPCADClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::PCA_D);
}
#ifdef USE_GSL
void WaveSpaceGui::OnRbtnSplineClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::SPLINE);
}
void WaveSpaceGui::OnRbtnSplineDClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::SPLINE_D);
}
void WaveSpaceGui::OnRbtnFourierClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::FOURIER);
}
void WaveSpaceGui::OnRbtnFourierDClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::FOURIER_D);
}
#endif
void WaveSpaceGui::OnRbtnWaveletsClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::WAVELET);
}
void WaveSpaceGui::OnRbtnWaveletsDClicked(bool)
{
    _wsgSpec->SetGuiTransform(wv::WAVELET_D);
}

void WaveSpaceGui::OnLsUTContentsNewIndex(int index)
{
    QListWidgetItem* item = ui->lsUTContents->item(index);
    OnLsUTContentsItemClicked(item, true);
}
void WaveSpaceGui::OnLsUTContentsItemClicked(QListWidgetItem* item, bool is_prog)
{
/*    int index = ui->lsUTContents->row(item);
    OnRbtnUserClicked(true);
    _wsgSpec->SetTransform( _wsgSpec->ActiveAxis(), wv::USER, index );
    _wsgSpec->SetWavesTransform(wv::USER, _wsgSpec->CurrentUserTid() );
    _wsgSpec->SetUserTCIndex(index);
    if (!is_prog)
        UpdateAll();*/
}
void WaveSpaceGui::OnSpnUserTransformValueChanged(int value, bool is_prog)
{
    _wsgSpec->SetUserTidIndex(value);
    UpdateUTContents();
    if (!is_prog)
        UpdateAll();
}

void WaveSpaceGui::OnBtnTemplatesClicked(bool)
{
    if (ui->btnTemplates->text() == "Average")
    {
        if ( !_cmed.ClusAccepted(wv::DEFAULT_CID) ) return;
        if (_needNewTemplates)
        {
            MakeTemplates();
            _needNewTemplates = false;
        }
        ui->wgtWaves->SetMode(WaveWidget::TEMPLATES);
        UpdateWavesWidget();
        ui->btnTemplates->setText("Individual");
    }
    else
    {
        ui->wgtWaves->SetMode(WaveWidget::WAVEFORMS);
        ui->btnTemplates->setText("Average");
    }
    ui->wgtWaves->update();
}

void WaveSpaceGui::OnBtnSetSamplesClicked(bool is_prog)
{
    _cluSpec->SetSelectedSamplesBegin( _wsgSpec->SamplesBegin() );
    _cluSpec->SetSelectedSamplesEnd( _wsgSpec->SamplesEnd() );

    if (!is_prog) ui->wgtWaves->update();
}

void WaveSpaceGui::OnBtnClusterClicked(bool)
{
    wv::CLUSTER_METHOD method = _cluSpec->ClusterMethod();
    if (_wsgSpec->GuiTransform()==wv::USER && method==wv::MANUAL)
    {
        Log("Sorry, manual clustering of user transform vectors is not allowed.  Replace it with "
            "the transform vector whence it was derived.");
        return;
    }

    _cluSpec->SetTransform();
    if ( !_cluSpec->ClusterParameters() )
        _cluSpec->SetClusterParameters( ClusterParamsBase::Create(method) );
    _cluSpec->ClusterParameters()->SetSampleRangePars();

    if ( _cluSpec->ClusterMethod()==wv::MANUAL )
    {
        FinishDrawline();
        ui->wgtWaves->update();
    }

    if (_cluSpec->Transform()==wv::USER && _cluSpec->TransformId()==wv::DEFAULT_TID)
    {
        wv::TRANSFORM_ID old_tid = _cluSpec->TransformId();
        int new_tid = _waveData->CreateUserTransform(old_tid);
        _cluSpec->SetTransformId(new_tid);
    }

    // *** Here is where we create the new cluster ID and entry in ClusterMediator
    InputFilePtr input_file = _cmed.GetInputFile(wv::DEFAULT_CID);
    short electrode = _cmed.GetElectrode(wv::DEFAULT_CID);
    wv::CLUSTER_ID cid = _cmed.IncrementCID();
    _cmed.AddInputFile(cid, input_file);
    _cmed.SetElectrode(cid, electrode);
    _cmed.AddClusterSpec(cid, _cluSpec->Clone());
    if (_cmed.HasResults(wv::DEFAULT_CID) && !_cmed.Accepted(wv::DEFAULT_CID).empty())
        _cluSpec->SetPriorAccepted( _wsgSpec->Accepted() );
    ClusterMgrPtr cmgr( new ClusterMgr( std::vector<wv::CLUSTER_ID>(1,cid) ) );
    _cmed.AddClusterMgr(cmgr);
    //Currently, if the clustering fails, these
    //entries will continue to hang around...
    cmgr->StartThreads();
}

void WaveSpaceGui::OnBtnResetClassificationClicked(bool)
{
    ResetClassification();
}

void WaveSpaceGui::OnRbtnManualClicked(bool)
{
    _cluSpec->SetClusterMethod(wv::MANUAL);
    _wsgSpec->SetClusBegin(DOUBLE_MAX);
    _wsgSpec->SetClusEnd(DOUBLE_MAX);
    SetDrawButtons(true);
        //Here and with the other cluster methods the parameters, so long
        //as they have been used, will get deleted in MainWindow::WaveGuiClosed
    ClearDrawLines();
    if (ui->rbtnUser->isChecked())
    emit SendParameters( new ManualPars() );
}
void WaveSpaceGui::OnRbtnAutoClassClicked(bool)
{
    _cluSpec->SetClusterMethod(wv::AUTOCLASS);
    _cluSpec->SetClusterParameters(nullptr);
    _wsgSpec->SetClusBegin(DOUBLE_MAX);
    _wsgSpec->SetClusEnd(DOUBLE_MAX);
    SetDrawButtons(false);
    ClearDrawLines();
    emit SendParameters( new AutoClassPars() );
}
void WaveSpaceGui::OnRbtnKMeansClicked(bool)
{
    _cluSpec->SetClusterMethod(wv::KMEANS);
    _cluSpec->SetClusterParameters(nullptr);
    _wsgSpec->SetClusBegin(DOUBLE_MAX);
    _wsgSpec->SetClusEnd(DOUBLE_MAX);
    SetDrawButtons(false);
    ClearDrawLines();
    emit SendParameters( new KMeansPars() );
}
void WaveSpaceGui::OnRbtnSPCClicked(bool)
{
    _cluSpec->SetClusterMethod(wv::SPC);
    _cluSpec->SetClusterParameters(nullptr);
    _wsgSpec->SetClusBegin(DOUBLE_MAX);
    _wsgSpec->SetClusEnd(DOUBLE_MAX);
    SetDrawButtons(false);
    ClearDrawLines();
    emit SendParameters( new SPCParameters() );
}
void WaveSpaceGui::OnRbtnKKClicked(bool)
{
    _cluSpec->SetClusterMethod(wv::KK);
    _cluSpec->SetClusterParameters(nullptr);
    _wsgSpec->SetClusBegin(DOUBLE_MAX);
    _wsgSpec->SetClusEnd(DOUBLE_MAX);
    SetDrawButtons(false);
    ClearDrawLines();
    emit SendParameters( new KKParameters() );
}

void WaveSpaceGui::on_edSliderValue_editingFinished()
{
    bool ok;
    int value = ui->edSliderValue->text().toDouble(&ok) + 0.5;
    if (ok)
    {
        ui->sldHistX->setValue(value);
        _wsgSpec->SetCurrentValue(value);
        if ( _wsgSpec->HistCursorMode()==WSGSpec::BAR ) DrawHistWindow();
        ui->wgtHistogram->update();
        UpdateWaves();
        UpdateWavesWidget();
    }
    else
        ui->edSliderValue->clear();
}
void WaveSpaceGui::OnCmbDrawLinesActivated(const QString& qtext, bool)
{
    std::string text = qtext.toStdString();
    StringVec plines = ut::Split<StringVec>(text, ",");
    boost::algorithm::trim( plines[0] );
    boost::algorithm::trim( plines[1] );
    StringVec pline_x = ut::Split<StringVec>( plines.at(0), " " ),
              pline_y = ut::Split<StringVec>( plines.at(1), " " );

    wv::TRANSFORM transform_x = wv::TransformEnumVal(pline_x.at(0).c_str()),
            transform_y = wv::TransformEnumVal(pline_y.at(0).c_str());
    int sample_x = atoi(pline_x.at(1).c_str()),
            sample_y = atoi(pline_y.at(1).c_str());

    _wsgSpec->SetTransform(WSGSpec::X, transform_x, sample_x);
    _wsgSpec->SetTransform(WSGSpec::Y, transform_y, sample_y);
        //No user transforms with Manual clustering
    _wsgSpec->SetGuiTransform(transform_y);
    ui->wgtHistogram->ResetScaleSpecVals();
    UpdateAll(false);
}
void WaveSpaceGui::OnBtnScrubClicked(bool)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::STOPPED:
        case HistWidget::DRAWING_CIRCLE:
        case HistWidget::DRAWING_ELLIPSE:
        case HistWidget::DRAWING_CENTER:
        case HistWidget::DRAWING_POLYLINE:
            FinishDrawline();
            ui->wgtHistogram->SetIsScrubbing(true);
            ui->wgtWaves->SetSelectMode(WaveWidget::SCRUB);
            ui->btnScrub->setText("Finished");
            break;
        case HistWidget::SCRUBBING:
            ui->wgtHistogram->SetIsScrubbing(false);
            ui->wgtWaves->SetSelectMode(WaveWidget::GRAB);
            ui->btnScrub->setText("Scrub");
            break;
    }
}
void WaveSpaceGui::OnBtnBiEllipseClicked(bool)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::DRAWING_ELLIPSE:
            if (_waitForEllipse)
            {
                FinishDrawline();
                ui->spnSD->hide();
                _waitForEllipse = false;
                ui->btnEllipse->setText("Ellipse");
            }
            else
            {
                FinishDrawline(false);
                _waitForEllipse = true;
                ui->spnSD->show();
                ui->btnEllipse->setText("Finish SD");
            }
            break;
        case HistWidget::DRAWING_CENTER:
        case HistWidget::DRAWING_CIRCLE:
        case HistWidget::DRAWING_POLYLINE:
        case HistWidget::SCRUBBING:
        case HistWidget::STOPPED:
            InitCircle();
            ui->btnEllipse->setText("Finish Circle");
            ui->wgtHistogram->StartEllipse();

            ui->btnResetClassification->setEnabled(false);
            break;
    }
}
void WaveSpaceGui::OnSpnSDValueChanged(wv::float_t sd, bool)
{
    BiEllipse* be = dynamic_cast<BiEllipse*>( _cluSpec->Circles().back()->GetDrawLine() );
    if (!be) throw BadGuiOp("WaveSpaceGui::OnSpnSDValueChanged", "Current DrawLine not an ellipse");
    be->SetSD(sd);
    UpdateHistWidget();
}
void WaveSpaceGui::OnBtnCenterClicked(bool)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::SCRUBBING:
            OnBtnScrubClicked(); //NB fall-through
        case HistWidget::STOPPED:
        {
            ui->wgtHistogram->StartCenter();
            ui->btnCenter->setText("Finish");
            ui->btnResetClassification->setEnabled(false);
            break;
        }
        case HistWidget::DRAWING_CENTER:
            FinishDrawline();
            break;
        case HistWidget::DRAWING_CIRCLE:
        case HistWidget::DRAWING_ELLIPSE:
        case HistWidget::DRAWING_POLYLINE:
            FinishDrawline();
            OnBtnCenterClicked();
            break;
    }
}
void WaveSpaceGui::OnBtnCircleClicked(bool)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::SCRUBBING:
            OnBtnScrubClicked(); //NB fall-through
        case HistWidget::STOPPED:
        {
            InitCircle();
            ui->wgtHistogram->StartCircle();
            ui->btnCircle->setText("Finish");

            ui->btnResetClassification->setEnabled(false);
            break;
        }
        case HistWidget::DRAWING_CIRCLE:
            FinishDrawline();
            break;
        case HistWidget::DRAWING_CENTER:
        case HistWidget::DRAWING_ELLIPSE:
        case HistWidget::DRAWING_POLYLINE:
            FinishDrawline();
            OnBtnCircleClicked();
            break;
    }
}
void WaveSpaceGui::OnBtnPolylineUpClicked(bool)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::SCRUBBING:
            OnBtnScrubClicked(); //NB fall-through
        case HistWidget::STOPPED:
            StartPolyline(Polyline::REJECT_UP);
            break;
        case HistWidget::DRAWING_POLYLINE:
            FinishDrawline();
            break;
        case HistWidget::DRAWING_CENTER:
        case HistWidget::DRAWING_ELLIPSE:
        case HistWidget::DRAWING_CIRCLE:
            FinishDrawline();
            OnBtnPolylineUpClicked();
            break;
    }
}
void WaveSpaceGui::OnBtnPolylineDownClicked(bool)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::SCRUBBING:
            OnBtnScrubClicked();
        case HistWidget::STOPPED:
            StartPolyline(Polyline::REJECT_DOWN);
            break;
        case HistWidget::DRAWING_POLYLINE:
            FinishDrawline();
            break;
        case HistWidget::DRAWING_CENTER:
        case HistWidget::DRAWING_CIRCLE:
        case HistWidget::DRAWING_ELLIPSE:
            FinishDrawline();
            OnBtnPolylineDownClicked();
            break;
    }
}
void WaveSpaceGui::OnBtnClearDrawLine(bool)
{
    FinishDrawline();
    ClearCurrentDrawLines();

    //Get the polyline names off the list
    CenterSpec center_spec(_wsgSpec->Transform(WSGSpec::X),
        _wsgSpec->Transform(WSGSpec::Y),
        _wsgSpec->Sample(WSGSpec::X),
        _wsgSpec->Sample(WSGSpec::Y), 0.0, 0.0);
        //***This is a hack to get the name easily
    std::string name = center_spec.Name();
    int i=0;
    while (i < ui->cmbDrawLines->count())
    {
        std::string pline = ui->cmbDrawLines->itemText(i).toStdString();
        if (pline == name)
            ui->cmbDrawLines->removeItem(i);
        else
            ++i;
    }

    UpdateHistWidget();
}

void WaveSpaceGui::MakeBAandColors(int num_records) //BA for boost::dynamic_bitset<>
{
    boost::dynamic_bitset<>& is_accepted = _wsgSpec->Accepted();
    is_accepted = boost::dynamic_bitset<>(num_records, false);
    _allColors = std::vector<QColor>(num_records, Qt::white);
    const wv_vector_int* clus_accepted = _cmed.ClusAccepted(wv::DEFAULT_CID);
    if (!clus_accepted) return;

    _clusters.clear();
    const boost::dynamic_bitset<>& is_toggled = _cluSpec->ToggledWaves();
    for (int i=0; i<num_records; ++i)
    {
        int c = wv_vector_int_get(clus_accepted, i);
        if ( std::find(_clusters.begin(), _clusters.end(), c)==_clusters.end() )
            _clusters.push_back(c);
        bool b = ((c==wv::ACCEPT_CLUSTER && !is_toggled.test(i))
                || (c==wv::REJECT_CLUSTER && is_toggled.test(i)));
        is_accepted.set(i, b);
        _allColors[i] = b ? Qt::green : Qt::white;
    }
    std::sort(_clusters.begin(), _clusters.end());
}
void WaveSpaceGui::StartPolyline(Polyline::REJECT_DIR reject_dir)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::SCRUBBING:
            OnBtnScrubClicked(true); //NB fall-through
        case HistWidget::STOPPED:
        {
            //Initialize the new polyline
            DrawLineBase::ID id = DrawLineBase::NewDrawLineID();
            Polyline* polyline = new Polyline(id, reject_dir);

            //Now initialize the corresponding PolylineSpec
            PLSpec* pl_spec = new PLSpec(_wsgSpec->Transform(WSGSpec::X),
                                         _wsgSpec->Transform(WSGSpec::Y),
                                         _wsgSpec->Sample(WSGSpec::X),
                                         _wsgSpec->Sample(WSGSpec::Y),
                                        polyline);

            _cluSpec->Polylines().push_back(pl_spec);
            ui->wgtHistogram->StartPolyline();
            switch (reject_dir)
            {
                case Polyline::REJECT_UP:
                    ui->btnPolylineUp->setText("Finish");
                    break;
                case Polyline::REJECT_DOWN:
                    ui->btnPolylineDown->setText("Finish");
                    break;
            }


            ui->btnResetClassification->setEnabled(false);
            break;
        }
        case HistWidget::DRAWING_CENTER:
        case HistWidget::DRAWING_POLYLINE:
        case HistWidget::DRAWING_CIRCLE:
        case HistWidget::DRAWING_ELLIPSE:
            FinishDrawline();
            StartPolyline(reject_dir);
            break;
    }
}
void WaveSpaceGui::ResetClassification()
{
    boost::dynamic_bitset<>& is_accepted = _wsgSpec->Accepted();
    is_accepted = boost::dynamic_bitset<>(_waveData->NumRecords(), false);
    _allColors = std::vector<QColor>(_waveData->NumRecords(), Qt::white);
    _cluSpec->InitToggledWaves();
    ui->btnScrub->setEnabled(false);

    ui->wgtWaves->SetMode(WaveWidget::WAVEFORMS);
    ui->btnTemplates->setText("Average");
    _cmed.SetClusAccepted(wv::DEFAULT_CID, nullptr);
    _cluSpec->SetSelectedSamplesBegin(0);
    int num_samples = _waveData->NumSamples( _wsgSpec->GuiTransform(), _wsgSpec->GuiTransformId() );
    _cluSpec->SetSelectedSamplesEnd(num_samples-1);

    ClearDrawLines();
    UpdateWavesWidget();
    UpdateHistogram(); //For updating the colors
    UpdateHistWidget();
}
void WaveSpaceGui::ResetDiskCursors()
{
    for (int i=0; i<WSGSpec::NUM_DISKS; ++i)
        if ( ui->wgtHistogram->DiskPixX(i)!=0 || ui->wgtHistogram->DiskPixY(i)!=0 )
        {
            int x = _scaleSpec->PixToValX( ui->wgtHistogram->DiskPixX(i) ),
                   y =  _scaleSpec->PixToValY( ui->wgtHistogram->DiskPixY(i) );
            wv::float_t hist_disk_size = (wv::float_t)_wsgSpec->HistDiskSize(i);
            wv::float_t xdist = _wsgSpec->Range(WSGSpec::X) * hist_disk_size/(wv::float_t)ui->wgtHistogram->width(),
                    ydist = _wsgSpec->Range(WSGSpec::Y) * hist_disk_size/(wv::float_t)ui->wgtHistogram->HistHeight();
            _wsgSpec->SetDiskCursor(i, x, y, xdist, ydist);
        }
}
void WaveSpaceGui::SafeUnload()
{
    _cmed.GetWaveData(wv::DEFAULT_CID)->Unload();
    _wsgSpec->ResetViews();
}
void WaveSpaceGui::ToggleWave(int scrub_index)
{
    _cluSpec->ToggledWaves().flip(scrub_index);
}

void WaveSpaceGui::SetHistCursor(HIST_CURSOR_MODE cursor_mode)
{
    _wsgSpec->SetHistCursorMode(cursor_mode);
    UpdateHistWidget();
    UpdateWaves();
    UpdateWavesWidget();
}
void WaveSpaceGui::SetHistAxis(WaveSpaceGuiSpec::AXIS axis)
{
    switch (axis)
    {
        case WSGSpec::X:
            ui->rbtnSetWavesFromXAxis->setChecked(true);
            break;
        case WSGSpec::Y:
            ui->rbtnSetWavesFromYAxis->setChecked(true);
            break;
    }

    _wsgSpec->SetTransform(axis, _wsgSpec->GuiTransform());
    _wsgSpec->SetActiveAxis(axis);
    _wsgSpec->SetWavesTransform(_wsgSpec->GuiTransform());
    _needNewTemplates = true;

    _cluSpec->SetSelectedSamplesBegin(0);
    int samples_end = _waveData->NumSamples( _wsgSpec->GuiTransform(), _wsgSpec->GuiTransformId() );
    _cluSpec->SetSelectedSamplesEnd(samples_end);
    ui->sldWaves->setRange(0, samples_end);
    ui->wgtHistogram->ResetScaleSpecVals();

    UpdateAll(true); //Need to reset for the histogram if nothing else--but also
        //if either the transform being replaced, or the current one, was wv::USER
        //we need to reset the samples.
    SafeUnload();
}
void WaveSpaceGui::SetDrawButtons(bool is_shown)
{
    if (is_shown)
    {
        ui->btnScrub->show();
        ui->btnCenter->show();
        ui->btnCircle->show();
        ui->btnEllipse->show();
        ui->btnPolylineUp->show();
        ui->btnPolylineDown->show();
        ui->btnClearDrawline->show();
        ui->cmbDrawLines->show();
        ui->line1->show();
        ui->line2->show();
        ui->line3->show();
    }
    else
    {
        ui->btnScrub->hide();
        ui->btnCenter->hide();
        ui->btnCircle->hide();
        ui->btnEllipse->hide();
        ui->spnSD->hide();
        ui->btnPolylineUp->hide();
        ui->btnPolylineDown->hide();
        ui->btnClearDrawline->hide();
        ui->cmbDrawLines->hide();
        ui->line1->hide();
        ui->line2->hide();
        ui->line3->hide();
    }
}
void WaveSpaceGui::SetMovieGui(bool is_shown, bool is_update)
{
    int num_children = ui->layMovie->count();
    for (int i=0; i<num_children; ++i)
        if (ui->layMovie->itemAt(i)->widget())
        {
            if (is_shown)
                ui->layMovie->itemAt(i)->widget()->show();
            else
                ui->layMovie->itemAt(i)->widget()->hide();
        }
    ui->layMovie->setEnabled(is_shown);
    if (is_update) UpdateHistWidget();
}
void WaveSpaceGui::SetUserControls(bool is_shown)
{
    if (is_shown)
    {
        ui->rbtnUser->setEnabled(true);
        ui->spnUserTransform->setMaximum( _waveData->NumUserTransforms()-1 );
        ui->spnUserTransform->setEnabled(true);
        ui->lsUTContents->setEnabled(true);
    }
    else
    {
        ui->rbtnUser->setEnabled(false);
        ui->spnUserTransform->setEnabled(false);
        ui->lsUTContents->setEnabled(false);
    }
}

void WaveSpaceGui::UpdateAll(bool reset_samples)
{
    //Which sample (in the waves widget) is chosen)
    if (reset_samples) _wsgSpec->SetSample(_wsgSpec->ActiveAxis(), 0);

    //Starting with the actual data values
    UpdateHistogram();
    if ( _wsgSpec->HistCursorMode()==WSGSpec::DISK) ResetDiskCursors();
    UpdateWaves();

    //Now update the gui, e.g. reposition sliders, redraw windows, etc.
    UpdateHistWidget(reset_samples); //Reset window
    UpdateWavesWidget(reset_samples);

    UpdateUTContents();
}

void WaveSpaceGui::UpdateHistogram()
{
#ifdef QT_DEBUG
    QTime tic;
    tic.start();
#endif

    if (_histAccept) { wv_histogram2d_free(_histAccept); _histAccept = nullptr; }
    if (_histUnaccept) { wv_histogram2d_free(_histUnaccept); _histUnaccept = nullptr; }
    if (_matAccept) { wv_matrix_int_free(_matAccept); _matAccept = nullptr; }
    if (_matUnaccept) { wv_matrix_int_free(_matUnaccept); _matUnaccept = nullptr; }

    //Now make sure the bin counts are within range.  The first time WaveSpaceGui is opened, the histogram
    //doesn't have full height and width.
    int hist_wd = qMax(ui->wgtHistogram->width(), DEFAULT_NUM_XBINS-1),
            hist_ht = qMax(ui->wgtHistogram->HistHeight(), DEFAULT_NUM_YBINS-1);
    _numXBins = qBound(10, _numXBins, hist_wd);
    _numYBins = qBound(10, _numYBins, hist_ht);

    //Now we can allocate and populate the matrices
    const wv::float_t delta = 0.00001; //To compensate for roundoff error and to ensure that the last
        //point is included in the histogram
    wv::float_t x_inc = delta*_wsgSpec->Range(WSGSpec::X),
            y_inc = delta*_wsgSpec->Range(WSGSpec::Y);
    wv::float_t min_x = _wsgSpec->Min(WSGSpec::X) - x_inc,
            max_x = _wsgSpec->Max(WSGSpec::X) + x_inc,
            min_y = _wsgSpec->Min(WSGSpec::Y) - y_inc,
            max_y = _wsgSpec->Max(WSGSpec::Y) + y_inc;

    //Copy the relevant transform data into vectors for the histogram.
    const wv_vector* xdata_vec = _wsgSpec->HistVector(WSGSpec::X),
            * ydata_vec = _wsgSpec->HistVector(WSGSpec::Y);



    //Create the histograms, in parallel
    const boost::dynamic_bitset<>& is_accepted = _wsgSpec->Accepted();
    HistSpec hist_spec(is_accepted, *_wsgSpec);
    hist_spec.num_bins_x = _numXBins;
    hist_spec.num_bins_y = _numYBins;
    hist_spec.min_x = min_x;
    hist_spec.min_y = min_y;
    hist_spec.max_x = max_x;
    hist_spec.max_y = max_y;
    hist_spec.xdata_vec = xdata_vec;
    hist_spec.ydata_vec = ydata_vec;
    if ( _wsgSpec->IsScatter() ) // ***This routine is not actually faster than just using histograms
        //for this!!! Lol! However the gain in speed in HistWidget::paintEvent seems to make up for it.
        //---But only for files with less spikes (<300,000)
    {
        _matAccept = wv_matrix_int_alloc(_numXBins, _numYBins);
        _matUnaccept = wv_matrix_int_alloc(_numXBins, _numYBins);
        wv_matrix_int_set_zero(_matAccept);
        wv_matrix_int_set_zero(_matUnaccept);
        ui->wgtHistogram->SetMatrices(_matAccept, _matUnaccept);
        HistScatterMgr mgr(ui->wgtHistogram, hist_spec);
        mgr.StartThreads();
        mgr.Wait();
        _matAccept = mgr.MatAccept();
        _matUnaccept = mgr.MatUnaccept();
    }
    else
    {
        HistThreadMgr mgr(hist_spec);
        mgr.StartThreads();
        mgr.Wait();
        _histAccept = mgr.HistAccept();
        _histUnaccept = mgr.HistUnaccept();
    }

#ifdef QT_DEBUG
//    qDebug() << "WaveSpaceGui::UpdateHistogram(), "
//             << (_wsgSpec->IsScatter() ? "scatter" : "histogram") << std::string(": ").c_str()
//             << _numXBins << ", " << _numYBins << ": " << tic.elapsed();
#endif

    if (xdata_vec->size==1)
    {
        _matAccept = _matUnaccept = NULL;
        _histAccept = _histUnaccept = NULL;
    }

    //Put it in the histogram widget, along with information about the actual data ranges
    _wsgSpec->SetWindowLimits( ui->edWindow->text().toInt() );
    if ( _wsgSpec->IsScatter() )
        ui->wgtHistogram->SetMatrices(_matAccept, _matUnaccept);
    else
        ui->wgtHistogram->SetHistograms(_histAccept, _histUnaccept);
}
void WaveSpaceGui::UpdateWaves()
{
    const wv_vector* vec = _wsgSpec->HistVector(WSGSpec::X); //It's X because there the slider is along
        //the X axis
    int num_records = vec->size;

    int num_shown = ui->edNumShown->text().toInt(),
            total_viewable = ui->edWindow->text().toInt(),
            step = (int)( (wv::float_t)total_viewable / (wv::float_t)num_shown + 0.5 );
    const boost::dynamic_bitset<>& window = _wsgSpec->Window();
    wv::float_t window_begin = -1.0,
            window_end = -1.0;
    if (_wsgSpec->HistLimitMode() == WSGSpec::TIMESTAMP)
    {
        window_begin = _wsgSpec->TimeBegin();
        window_end = _wsgSpec->TimeEnd();
    }

    if (ui->btnTemplates->text()=="Individual") MakeTemplates();

    step = std::max(step, 1);
    boost::dynamic_bitset<>& is_wave_shown = _wsgSpec->WaveShown();
    is_wave_shown.reset();
    switch ( _wsgSpec->HistCursorMode() )
    {
        case WSGSpec::BAR:
            if (_wsgSpec->HistLimitMode()==WSGSpec::TRANSFORM)
            {
                bool no_step = total_viewable <= num_shown;
                for (int i=0, ct=0; i<num_records; ++i)
                    if ( window.test(i) && (no_step || ct++%step==0) )
                        is_wave_shown.set(i,true);
            }
            else //Spec::TIMESTAMP
            {
                bool no_step = total_viewable <= num_shown;
                const wv_matrix* timestamps = _wsgSpec->Timestamps().get();
                for (int i=0, ct=0; i<num_records; ++i)
                {
                    wv::float_t val = wv_matrix_get(timestamps, i, 0);
                    if (val>window_begin && val<=window_end && (no_step || ct++%step==0))
                        is_wave_shown.set(i,true);
                }
            }
            break;
        case WSGSpec::DISK:
        {
            const boost::dynamic_bitset<>* disk_cursors = _wsgSpec->DiskCursors();
            const QColor* disk_colors = ui->wgtHistogram->DiskColors();
            for (int j=0; j<WSGSpec::NUM_DISKS; ++j)
            {
                const boost::dynamic_bitset<>& disk_cursor = disk_cursors[j];
                const QColor& disk_color = disk_colors[j];
                int disk_ct = disk_cursor.count();
                if (disk_ct==0) continue;
                bool no_step = disk_ct <= num_shown;
                if (!no_step)
                    step = (int)((wv::float_t)disk_ct / (wv::float_t)num_shown + 0.5);
                if (_wsgSpec->HistLimitMode()==WSGSpec::TRANSFORM)
                {
                    for (int i=0, ct=0; i<num_records; ++i)
                        if (disk_cursors[j].test(i) && (no_step || ct++%step==0))
                        {
                            is_wave_shown.set(i,true);
                            _allColors[i] = disk_color; //Accepted/non-accepted taken care of in WaveWidget
                        }
                }
                else //Spec::TIMESTAMP
                {
                    const wv_matrix* timestamps = _wsgSpec->Timestamps().get();
                    for (int i=0, ct=0; i<num_records; ++i)
                    {
                        wv::float_t val = wv_matrix_get(timestamps, i, 0);
                        if (disk_cursors[j].test(i) && val>window_begin && val<=window_end
                                && (no_step || ct++%step==0))
                        {
                            is_wave_shown.set(i,true);
                            _allColors[i] = disk_color; //Accepted/non-accepted taken care of in WaveWidget
                        }
                    }
                }
            }
            break;
        }
    }
}

void WaveSpaceGui::UpdateTimeSlider()
{
    int window_size = ui->edWindow->text().toInt();
    _wsgSpec->SetWindowLimits(window_size);
    UpdateWavesWidget();
    UpdateWaves();
}
void WaveSpaceGui::UpdateHistWidget(bool reset_window)
{
    //Update which sliders are enabled
    if ( _wsgSpec->HistLimitMode()==WSGSpec::TRANSFORM)
    {
        if ( _wsgSpec->HistCursorMode()==WSGSpec::BAR )
            ui->sldHistX->setEnabled(true);
        else
            ui->sldHistX->setEnabled(false);
    }
    else
        ui->sldHistX->setEnabled(false);

    //Now the yellow selection window, after resetting the slider    
    if ( _wsgSpec->HistCursorMode()==WSGSpec::BAR ) DrawHistWindow(reset_window);

    //Axis x and y labels
    std::string xlabel, ylabel;
    wv::TRANSFORM transform_x = _wsgSpec->Transform(WSGSpec::X),
            transform_y = _wsgSpec->Transform(WSGSpec::Y);
    int sample_x = _wsgSpec->Sample(WSGSpec::X),
            sample_y = _wsgSpec->Sample(WSGSpec::Y);
    if (transform_x==wv::USER)
    {
        wv::TRANSFORM_ID tid = _wsgSpec->TransformId(WSGSpec::X);
        std::string tr = wv::TransformName( _waveData->UserTransform(tid, sample_x) );
        int sample = _waveData->UserSample(tid, sample_x);
        xlabel = "User: " + tr + ", sample " + std::to_string(sample);
    }
    else
        xlabel = wv::TransformName(transform_x) + ", sample " + std::to_string(sample_x);
    if (transform_y==wv::USER)
    {
        wv::TRANSFORM_ID tid = _wsgSpec->TransformId(WSGSpec::Y);
        std::string tr = wv::TransformName( _waveData->UserTransform(tid, sample_y) );
        int sample = _waveData->UserSample(tid, sample_y);
        ylabel = "User: " + tr + ", sample " + std::to_string(sample);
    }
    else
        ylabel = wv::TransformName(transform_y) + ", sample " + std::to_string(sample_y);
    AxisLabel(WSGSpec::X)->SetText(xlabel,WSGSpec::X);
    AxisLabel(WSGSpec::Y)->SetText(ylabel,WSGSpec::Y);

    //Now reset slider/bar properties and redraw the window_cid,
    ui->sldHistX->setRange( _scaleSpec->MinValX()+0.5, _scaleSpec->MaxValX()+0.5 );
    ui->wgtHistogram->update();
}
void WaveSpaceGui::UpdateWavesWidget(bool reset_samples) const
{
    WSGSpec::AXIS active_axis = _wsgSpec->ActiveAxis();
    MatCPtr waves = _waveData->Waveforms(_wsgSpec->ActiveTransform(), _wsgSpec->TransformId(active_axis));

    //Slider
    ui->sldWaves->setEnabled( waves->size2>1 && !ui->rbtnSetWavesFromRaw->isChecked());
    if (reset_samples)
    {
        int num_samples = waves->size2;
        ui->sldWaves->setMinimum(0);
        _cluSpec->SetSelectedSamplesBegin(0);
        ui->sldWaves->setMaximum(num_samples-1);
        _cluSpec->SetSelectedSamplesEnd(num_samples);
    }
    ui->sldWaves->setValue( ui->rbtnSetWavesFromRaw->isChecked()
                            ? 0
                            : _wsgSpec->Sample( _wsgSpec->ActiveAxis() ) );
    if ( ui->rbtnSetWavesFromRaw->isChecked() )
        _cluSpec->SetSelectedSamplesBegin(-1);
    else if (_cluSpec->SelectedSamplesBegin()==-1)
        _cluSpec->SetSelectedSamplesBegin(-1);

    //Draw yellow scan line
    DrawScanline();

    //Now redraw the widget appropriately
    ui->wgtWaves->SetWaves(wv::DEFAULT_CID, _wsgSpec->WavesTransform(), _wsgSpec->TransformId(active_axis),
                           &_allColors, reset_samples);
    ui->wgtWaves->SetTemplates(wv::DEFAULT_CID);
    if (reset_samples) ui->wgtWaves->ResetSampleRange();
        //This must come *after* thet call to SetWaves, so that _allWaves has been set appropriately
    ui->wgtWaves->update();

    //Send the selected waves to _waveGui, if there is one
    if (_cmed.CurrentWaveGui()!=wv::DEFAULT_CID)
        emit UpdateWavesShown();
}
void WaveSpaceGui::UpdateUTContents()
{
    if (_waveData->NumUserTransforms()==0) return;
    static bool is_updating = false;
    if (is_updating) return;
    is_updating = true;

    ui->lsUTContents->clear();
    if (_wsgSpec->UserTCIndex()==-1) _wsgSpec->SetUserTCIndex(0);
    if (_wsgSpec->UserTidIndex()==-1)
    {
        _wsgSpec->SetUserTidIndex(0);
        ui->spnUserTransform->setValue(0);
    }
    const UTS& ut_spec = _waveData->UTSpecAt( _wsgSpec->UserTidIndex() );
    wv::TRANSFORM_ID tid = ut_spec.ID();
    int num_specs = ut_spec.NumSpecs();
    for (int i=0; i<num_specs; ++i)
    {
        std::string label = _waveData->TransformLabel(wv::USER, i, tid);
        if (label.substr(0, 6)=="User: ") label = label.substr(6);
        ui->lsUTContents->addItem(label.c_str());
    }
    is_updating = false;
}

void WaveSpaceGui::MakeTemplates()
{
    const wv_vector_int* clus_accepted = _cmed.ClusAccepted(wv::DEFAULT_CID);
    if (!clus_accepted) return;
    _wsgSpec->MakeTemplates(_clusters, clus_accepted, wv::DEFAULT_CID);
}
void WaveSpaceGui::FinishDrawline(bool stop)
{
    switch (ui->wgtHistogram->DrawlineState())
    {
        case HistWidget::DRAWING_POLYLINE:
        {
            bool is_ok = ui->wgtHistogram->FinishedDrawLine();
            if (is_ok)
            {
                std::string name
                        = const_cast<ClusterSpec*>(_cluSpec.get())->Polylines().back()->Name();
                ui->cmbDrawLines->addItem(name.c_str());
                UpdateHistWidget();
            }
            break;
        }
        case HistWidget::DRAWING_CIRCLE:
        {
            bool is_ok = ui->wgtHistogram->FinishedDrawLine();
            if (is_ok)
            {
                std::string name
                        = const_cast<ClusterSpec*>(_cluSpec.get())->Circles().back()->Name();
                ui->cmbDrawLines->addItem(name.c_str());
                UpdateHistWidget();
            }
            break;
        }
        case HistWidget::DRAWING_ELLIPSE:
        {
            bool is_ok = ui->wgtHistogram->FinishedDrawLine(stop);
            if (is_ok)
            {
                std::string name
                        = const_cast<ClusterSpec*>(_cluSpec.get())->Circles().back()->Name();
                ui->cmbDrawLines->addItem(name.c_str());
                UpdateHistWidget();
            }
            break;
        }
        case HistWidget::DRAWING_CENTER:
            ui->wgtHistogram->FinishedDrawLine();
            break;
        case HistWidget::SCRUBBING:
        case HistWidget::STOPPED:
            break;
    }
    ResetDrawGui();
    ui->btnResetClassification->setEnabled(true);
}
void WaveSpaceGui::ResetDrawGui()
{
    ui->btnPolylineUp->setText("Polyline Up");
    ui->btnPolylineDown->setText("Polyline Down");
    ui->btnCircle->setText("Circle");
    ui->btnEllipse->setText("Ellipse");
    ui->btnCenter->setText("Centers");
    ui->spnSD->hide();
    _waitForEllipse = false;
}

void WaveSpaceGui::ClearDrawLines()
{
    _cluSpec->DeleteDrawLines();
    ui->cmbDrawLines->clear();
    if (ui->wgtHistogram->DrawlineState()==HistWidget::SCRUBBING)
        OnBtnScrubClicked();
}
void WaveSpaceGui::ClearCurrentDrawLines()
{
    std::vector<PLSpec*>& polylines = _cluSpec->Polylines();
    std::vector<CSpec*>& circles = _cluSpec->Circles();
    wv::TRANSFORM tx = _wsgSpec->Transform(WSGSpec::X),
            ty = _wsgSpec->Transform(WSGSpec::Y);
    int sx = _wsgSpec->Sample(WSGSpec::X),
            sy = _wsgSpec->Sample(WSGSpec::Y);

    auto itp = polylines.begin();
    while (itp != polylines.end())
        if ((*itp)->sample_x == sx && (*itp)->sample_y == sy
                && (*itp)->transform_x == tx && (*itp)->transform_y == ty)
            itp = polylines.erase(itp);
        else
            ++itp;

    auto itc = circles.begin();
    while (itc != circles.end())
        if ((*itc)->sample_x == sx && (*itc)->sample_y == sy
                && (*itc)->transform_x == tx && (*itc)->transform_y == ty)
            itc = circles.erase(itc);
        else
            ++itc;

    _cluSpec->ClearCenters();
}


void WaveSpaceGui::DrawScanline() const
{
    int value = _wsgSpec->Sample( _wsgSpec->ActiveAxis() );
    ui->wgtWaves->SetScanline(value);
        //You actually can't use ui->sldWaves->value() here, when this is called as result
        //of slider movement, it's still on the old value
}
void WaveSpaceGui::DrawHistWindow(bool reset_window) const
{
    int win_xbegin = -1,
            win_width = -1,
            win_ybegin = -1,
            win_height = -1;
                //The above 4 are in pixels

    if (reset_window)
        win_width = 0;
    else
    {
        wv::float_t window_begin_val = _wsgSpec->WindowBegin(),
                window_end_val = _wsgSpec->WindowEnd();
                    //The last two are determined by active axis, in *value* coordinates

        win_xbegin = _scaleSpec->ValToPixX(window_begin_val);
        int win_xend = _scaleSpec->ValToPixX(window_end_val);

        win_width = win_xend - win_xbegin;
        win_width = qMax(win_width,1);
    }
    win_ybegin = 0;
    win_height = ui->wgtHistogram->HistHeight();

    QRect window(win_xbegin, win_ybegin, win_width, win_height);
    ui->wgtHistogram->SetWindow(window);
}
void WaveSpaceGui::InitCircle()
{
    DrawLineBase::ID id = DrawLineBase::NewDrawLineID();
    Circle* circle = new Circle(id);

    //Now initialize the corresponding CircleSpec
    CSpec* c_spec = new CSpec(_wsgSpec->Transform(WSGSpec::X),
                              _wsgSpec->Transform(WSGSpec::Y),
                              _wsgSpec->Sample(WSGSpec::X),
                              _wsgSpec->Sample(WSGSpec::Y),
                              circle);

    _cluSpec->Circles().push_back(c_spec);
}
