#ifndef WAVEWIDGET_H
#define WAVEWIDGET_H

#include "../../../globals/log.h"
#include "../../../memrep/classification.h"
#include "../../../memrep/wavetemplate.h"
#include "../../../memrep/wavedata.h"

class WaveWidget : public QWidget
{
    Q_OBJECT

    public:
        WaveWidget(QWidget* parent);

        enum SELECT_MODE
        {
            GRAB,
            SCRUB
        };

        enum MODE
        {
            WAVEFORMS,
            TEMPLATES
        };

        enum AXIS_LIMIT_MODE
        {
            ABSOLUTE,
            RELATIVE
        };

        void ResetSampleRange();
        void SetSelectMode(SELECT_MODE select_mode) { _selectMode = select_mode; }
        void SetMode(MODE mode) { _mode = mode; }
        void SetBackground(QColor background);
        void SetTemplates(wv::CLUSTER_ID cid,
                          const std::map<int, bool>* is_template_shown = NULL, int level_idx = -1);
        virtual void SetWaves(wv::CLUSTER_ID cid, wv::TRANSFORM trans, wv::TRANSFORM_ID tid,
                      const std::vector<QColor>* all_colors, bool reset_samples = false);
            //Arguably the computation of which waves of the total are shown should be done in the parent
            //gui, but then the submatrix would have to be stored there, redundantly storing waves in two
            //member matrices.  This way seems better.
        void SetAxisLimitMode(AXIS_LIMIT_MODE axis_limit_mode) { _axisLimitMode = axis_limit_mode; }
        void SetAxisAbsoluteYRange(wv::float_t min, wv::float_t max);
        virtual void SetSamplesBegin(int samples_begin) = 0;
        virtual void SetSamplesEnd(int samples_end) = 0;
        void SetUseDashes(bool use_dashes) { _useDashes = use_dashes; }

        wv::float_t AbsoluteYMax() const { return _absoluteYMax; }
        wv::float_t AbsoluteYMin() const { return _absoluteYMin; }
        MatCPtr AllWaves() const;
        virtual bool CanDraw() const { return (bool)AllWaves(); }
        wv::CLUSTER_ID CID() const { return _cid; }
        int CurrentMouseX() const { return _currentMouseX; }
        int CurrentMouseY() const { return _currentMouseY; }
        wv::float_t CurrentYMax() const { return _currentYMax; }
        wv::float_t CurrentYMin() const { return _currentYMin; }
        MODE Mode() const { return _mode; }
        virtual int MSize1() const { return AllWaves()->size1; }
        virtual int MSize2() const { return AllWaves()->size2; }
        int NumSamples() const { return AllWaves()->size2; }
        virtual int SamplesBegin() const = 0;
        virtual int SamplesEnd() const = 0;
        int ScrubIndex() const { return _scrubIndex; }
        SELECT_MODE SelectMode() const { return _selectMode; }

    public slots:
        void ReceiveScrubIndex(int scrub_index);

    signals:
        void RangeChange(int begin, int end); //For the slider
        void SendScrubIndex(int scrub_index);

    protected:
        virtual void paintEvent(QPaintEvent*) override;

        void DrawTemplates();
        void DrawWaves();
        void FindScales(wv::float_t ymin, wv::float_t ymax, int sample_range,
                        wv::float_t& ht_scale, wv::float_t& wd_scale) const;
            //The multipliers which are returned convert data values into pixel values.

        void SetSampleRange(int begin, int end, bool is_restrictive=false); //So basically this is for ScanWaveWidget
            //is_restrictive determines whether, in the situation where one of begin/end is out of range,
            //the resizing is canceled, or the valid change is still made.
        void SetScrubIndex(int scrub_index) { _scrubIndex = scrub_index; }
        virtual void SetWavesShown(wv_matrix* waves_shown);
        void SetMousePos(int x, int y);

        WaveData* GetWaveData() { return _waveData; }
        virtual int NumWavesShown() const;
        virtual const boost::dynamic_bitset<>& WaveShown() const;

    private:
        wv::float_t _absoluteYMin, _absoluteYMax;
            //Need to make these members so that plotting polylines can use them
        const std::vector<QColor>* _allColors;
            //Pointers instead of references because they're not initialized at construction
        AXIS_LIMIT_MODE _axisLimitMode;
        wv::CLUSTER_ID _cid;
        int _currentMouseX, _currentMouseY;
        wv::float_t _currentYMin, _currentYMax;
            //Need to make these members so that plotting polylines can use them
        const std::map<int, bool>* _isTemplateShown;  //For templates; which templates to show.
        MODE _mode;
        int _scrubIndex;
        SELECT_MODE _selectMode;
        std::vector<QColor> _templateColors;
        wv::TRANSFORM_ID _tid;
        wv::TRANSFORM _transform;
        bool _useDashes;
            //This is really just for ScanWaveWidget, but needs to be done here
        WaveData* _waveData;
        const WaveTemplate* _waveTemplates;
};

#endif // WAVEWIDGET_H
