#include "polyline.h"

#ifndef USE_GSL
const int Polyline::LUT_LENGTH = 64*1024;
#endif

Polyline::Polyline(ID id, REJECT_DIR reject_dir)
    : DrawLineBase(id), _rejectDir(reject_dir)
#ifdef USE_GSL
        , _accel(nullptr), _spline(nullptr)
#else
        , _lut(nullptr)
#endif
{
}
Polyline::Polyline() : DrawLineBase()
#ifdef USE_GSL
      , _accel(nullptr), _spline(nullptr)
#else
      , _lut(nullptr)
#endif
{
}
Polyline::Polyline(const Polyline& polyline) : DrawLineBase(polyline),
      _rejectDir(polyline._rejectDir)
#ifdef USE_GSL
        , _accel(nullptr), _spline(nullptr)
#else
        , _lut(nullptr)
#endif
{
    MakeSpline();
}
Polyline& Polyline::operator=(const Polyline& polyline)
{
    if (this != &polyline)
    {
        DrawLineBase::operator=(polyline);
        _rejectDir = polyline._rejectDir;
#ifdef USE_GSL
        _spline = nullptr;
        _accel = nullptr;
#else
        _lut = nullptr;
#endif
        MakeSpline();
    }
    return *this;
}
Polyline::~Polyline()
{
#ifdef USE_GSL
    if (_accel) gsl_interp_accel_free(_accel);
    if (_spline) gsl_spline_free(_spline);
#else
    if (_lut) delete[] _lut;
#endif
}

void Polyline::AddPoint(QPointF point)
{
    if (_length==DRAWLINE_MAXLEN) throw BadDrawLine("Polyline::AddPoint", "Too many points");

    //The following is required in order to make sure the points are ordered by x value, needed
    //for the spline functino
    int i;
    float xval = point.x();

    //Find the first x value smaller than the new one, and get the array index--the new value
    //will go in after it.
    i=0;
    while (i<_length)
    {
        if ( fabs(xval - _points[i].x()) <= std::numeric_limits<float>::epsilon() )
            return;
        if (xval > _points[i].x())
        {
            ++i;
            continue;
        }
        break;
    }

    //Shift the points 1 to the right in the array, to make room for the new value
    for (int j=_length; j>i; --j)
    {
        _points[j].setX( _points[j-1].x() );
        _points[j].setY( _points[j-1].y() );
    }

    //Now put the new values in their place.
    _points[i].setX(xval);
    _points[i].setY( point.y() );

    ++_length;
    MakeSpline();
}

wv::float_t Polyline::Value(wv::float_t x) const
{
    if (_length<2 || x<_points[0].x() || x>_points[_length-1].x())
        return std::numeric_limits<wv::float_t>::quiet_NaN();
#ifdef USE_GSL
    return gsl_spline_eval(_spline, x, _accel);
#else
    int idx = LUT_LENGTH * (x - _points[0].x()) / (_points[_length-1].x() - _points[0].x()) + 0.5;
    return _lut[idx];
#endif
}

void Polyline::MakeSpline()
{
    if (_length<2) return;

#ifndef Q_OS_WIN
//    wv::float_t X[DRAWLINE_MAXLEN], Y[DRAWLINE_MAXLEN];
    wv::float_t X[DRAWLINE_MAXLEN], Y[DRAWLINE_MAXLEN];
#else
    wv::float_t X[2048], Y[2048];
#endif
    for (int i=0; i<_length; ++i)
    {
        X[i] = (wv::float_t)_points[i].x();
        Y[i] = (wv::float_t)_points[i].y();
    }

#ifdef USE_GSL
    _spline = gsl_spline_alloc(gsl_interp_linear, _length);
    _accel = gsl_interp_accel_alloc();

    gsl_spline_init(_spline, X, Y, _length);
#else
    if (_lut) delete[] _lut;
    _lut = new wv::float_t[LUT_LENGTH];
    int* idx = new int[LUT_LENGTH],
            * idxp1 = new int[LUT_LENGTH];
    wv::float_t* alpha = new wv::float_t[LUT_LENGTH];

    msl::InitLinInterp(LUT_LENGTH, _length, X, idx, alpha, idxp1);
    msl::vector_view<wv::float_t> vlut = msl::vector_view_array(_lut, LUT_LENGTH),
            vY = msl::vector_view_array(Y, _length);
    msl::LinInterp(&vlut.vector, &vY.vector, idx, alpha, idxp1);

    delete[] idx;
    delete[] alpha;
    delete[] idxp1;
#endif
}
