#ifndef CLUSTERSPEC_H
#define CLUSTERSPEC_H

#include "../../globals/globals.h"
#include "../../memrep/drawing/polyline.h"
#include "../../memrep/drawing/circle.h"
#include "../../memrep/drawing/centerspec.h"
#include "../../memrep/wavedata.h"
#include "wavespaceguispec.h"

//This is the basic state of the currently available data. It is everything needed to run a
//cluster analysis and nothing else (BatchMgr).  It is also used by WaveSpaceGui and possibly WaveGui properly

//Note that the clustering here *does not care where it came from*--InputFile, electrode should not be
//stored here

class ClusterSpec;
typedef std::shared_ptr<ClusterSpec> CluSpecPtr;

//All non-const methods are private here, accessible only to WaveSpaceGui--but which clu_spec is *current* can be
//set by anybody
// ***Only*** setter/getter, add/remove methods in this and all specs classes
class ClusterSpec
{
    friend class MainWindow;
    friend class WaveSpaceGui;
    friend class ClusterMediator;
    friend class ClusterTO; //For SetTransformId

    public:
        ~ClusterSpec();

        void InitToggledWaves(); //Assumes _waves has already been set
        void InitWaves(wv::CLUSTER_ID cid); //Waves need to be explicitly set when the clustering was generated during batch processing

        CluSpecPtr Clone() const; // Creates an identical copy
        CluSpecPtr MintNew() const; //Only copies a few features--just those needed for clustering--to the new ClusterSpec
            //Necessary for batch processing

        void AddCenter(const CenterSpec* center) { _centers.push_back(center); }
        void ClearCenters() { _centers.clear(); }
        void SetClusterParameters(ClusterParamsPtr parameters) { _clusterParameters = parameters; }
            //Needs to be used by batch
        void SetWaveData(WaveData* wave_data);

        const std::vector<const CenterSpec*>& Centers() const { return _centers; }
        std::vector<CSpec*>& Circles() { return _circles; }
        const std::vector<CSpec*>& Circles() const { return _circles; }
        wv::CLUSTER_METHOD ClusterMethod() const { return _clusterMethod; }
        ClusterParamsPtr ClusterParameters() { return _clusterParameters; }
        WSGSpec& GetWSGSpec() { return _wsgSpec; }
        int NumSamples() const;
        std::vector<PLSpec*>& Polylines() { return _polylines; }
        const std::vector<PLSpec*>& Polylines() const { return _polylines; }
        const boost::dynamic_bitset<> PriorAccepted() const { return _priorAccepted; }
        int SelectedSamplesBegin() const { return _selectedSamplesBegin; }
        int SelectedSamplesEnd() const { return _selectedSamplesEnd; }
        wv::TRANSFORM Transform() const { return _transform; }
        wv::TRANSFORM_ID TransformId() const { return _tid; }
        const boost::dynamic_bitset<>& ToggledWaves() const { return _toggledWaves; }
        MatCPtr Waves() const;

    private:
        ClusterSpec();
#ifndef Q_OS_WIN
        ClusterSpec(const ClusterSpec& cluster_specs) = delete;
        ClusterSpec operator=(const ClusterSpec& cluster_specs) = delete;
#else
        ClusterSpec(const ClusterSpec& cluster_specs);
        ClusterSpec operator=(const ClusterSpec& cluster_specs);
#endif

        void DeleteDrawLines();

        void SetClusterMethod(wv::CLUSTER_METHOD cluster_method) { _clusterMethod = cluster_method; }
        void SetPriorAccepted(const boost::dynamic_bitset<>& accepted); //Non-trivial, xors with toggled
        void SetSelectedSamplesBegin(int selected_samples_begin)
            { _selectedSamplesBegin = selected_samples_begin; }
        void SetSelectedSamplesEnd(int selected_samples_end)
            { _selectedSamplesEnd = selected_samples_end; }
        void SetTransform(); //From the current WSGSpec active axis
        void SetTransform(wv::TRANSFORM transform, wv::TRANSFORM_ID tid = wv::DEFAULT_TID);
        void SetTransformId(wv::TRANSFORM_ID tid) { _tid = tid; }

        boost::dynamic_bitset<>& ToggledWaves() { return _toggledWaves; }

        std::vector<const CenterSpec*> _centers;
        std::vector<CSpec*> _circles;
        wv::CLUSTER_METHOD _clusterMethod;
        ClusterParamsPtr _clusterParameters;
        int _selectedSamplesBegin, _selectedSamplesEnd;
        //The range of samples used in the analysis, *not* necessarily the
            //same as the visible range.
        std::vector<PLSpec*> _polylines;
        boost::dynamic_bitset<> _priorAccepted; //For manual clustering; the XOR of whats accepted from the
            //previous clustering and any  _toggledWaves
        wv::TRANSFORM_ID _tid; //wv::DEFAULT_TID (default) unless the transform is wv::USER, in which case it's unique
        boost::dynamic_bitset<> _toggledWaves;
        wv::TRANSFORM _transform;
        WaveData* _waveData;
        WSGSpec _wsgSpec; //Needed when a particular clustering gets displayed in WaveSpaceGuiSpec.
};

#endif // CLUSTERSPEC_H
